import re
import logging
from typing import Any, Dict, List, Callable, Optional

import betterproto
from betterproto.lib.google.protobuf import Value, ListValue

import qm
from qm.grpc import qua
from qm.exceptions import QmQuaException
from qm.grpc.qua_config import QuaConfig
from qm.utils.protobuf_utils import Node
from qm.serialization.qua_node_visitor import QuaNodeVisitor
from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor
from qm.grpc.qua import (
    QuaProgram,
    QuaProgramType,
    QuaProgramScript,
    QuaProgramIfStatement,
    QuaProgramAnyStatement,
    QuaProgramForStatement,
    QuaProgramPlayStatement,
    QuaProgramWaitStatement,
    QuaProgramAlignStatement,
    QuaProgramPauseStatement,
    QuaProgramVarDeclaration,
    QuaProgramBinaryExpression,
    QuaProgramForEachStatement,
    QuaProgramMeasureStatement,
    QuaProgramVarRefExpression,
    QuaProgramLiteralExpression,
    QuaProgramZRotationStatement,
    QuaProgramAnyScalarExpression,
    QuaProgramAssignmentStatement,
    QuaProgramRampToZeroStatement,
    QuaProgramResetFrameStatement,
    QuaProgramResetPhaseStatement,
    QuaProgramSetDcOffsetStatement,
    QuaProgramStatementsCollection,
    QuaProgramArrayVarRefExpression,
    QuaProgramStrictTimingStatement,
    QuaProgramWaitForTriggerStatement,
    QuaProgramUpdateFrequencyStatement,
    QuaProgramResetGlobalPhaseStatement,
    QuaProgramUpdateCorrectionStatement,
    QuaProgramFastFrameRotationStatement,
    QuaProgramAdvanceInputStreamStatement,
    QuaProgramWaitForTriggerStatementElementOutput,
)

logger = logging.getLogger(__name__)


class QuaSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self._indent = 0
        self._lines: List[str] = []
        self.tags: List[str] = []

    def out(self) -> str:
        return "\n".join(["from qm import CompilerOptionArguments", "from qm.qua import *", ""] + self._lines + [""])

    def _out_lines(self) -> List[str]:
        return self._lines

    def _default_enter(self, node: Node) -> bool:
        if not isinstance(node, tuple(_dont_print)):
            logger.info(f"entering {type(node).__module__}.{type(node).__name__}")
        statement: Optional[Callable[[Node], str]] = _statements.get(type(node), None)  # type: ignore[assignment]

        if statement is not None:
            line = statement(node)
            self._line(line)

        block: Optional[Callable[[Node], str]] = _blocks.get(type(node), None)  # type: ignore[assignment]
        if block is not None:
            self._enter_block(block(node))
        return statement is None

    @staticmethod
    def _search_auto_added_stream(values: List[Value]) -> bool:
        v2 = values[2]
        is_auto_added_result = isinstance(v2, Value) and betterproto.which_one_of(v2, "kind") == (
            "string_value",
            "auto",
        )
        return is_auto_added_result

    def _fix_legacy_save(self, sp_line: str, node: ListValue) -> None:
        split_sp_line = sp_line.split(".")
        trace_name = split_sp_line[0]
        save_name = split_sp_line[-1].split('"')[1]
        save_name = re.sub(r"_input\d", "", save_name)
        is_auto_added_stream = self._search_auto_added_stream(node.values)
        if is_auto_added_stream:
            self._lines.pop()
            line_to_remove_index = None
            for i in range(len(self._lines)):
                if self._lines[i].find(f"{trace_name} = declare_stream") > 0:
                    line_to_remove_index = i
                self._lines[i] = self._lines[i].replace(trace_name, f'"{save_name}"')
            if line_to_remove_index:
                self._lines.pop(line_to_remove_index)

    def enter_qm_grpc_qua_QuaProgramCompilerOptions(self, node: qua.QuaProgramCompilerOptions):
        options_list = []
        if node.strict:
            options_list.append(f"strict={node.strict}")
        if len(node.flags) > 0:
            options_list.append(f"flags={node.flags}")
        options_string = ", ".join(options_list)
        if len(options_string) > 0:
            self._lines.insert(0, f"compiler_options = CompilerOptionArguments({options_string})\n")

        return False

    def enter_qm_grpc_qua_QuaResultAnalysis(self, node: qua.QuaResultAnalysis) -> bool:
        if len(node.model) > 0:
            self._enter_block("with stream_processing():")
            return True
        else:
            return False

    def enter_qm_grpc_qua_QuaProgramSaveStatement(self, node: qua.QuaProgramSaveStatement) -> bool:
        if node.tag not in self.tags:
            self._line(f"{node.tag} = declare_stream()")
            self.tags.append(node.tag)
        save_line = f"save({ExpressionSerializingVisitor.serialize(node.source)}, {node.tag})"
        self._line(save_line)
        return False

    def enter_qm_grpc_qua_QuaProgramVarDeclaration(self, node: qua.QuaProgramVarDeclaration) -> bool:
        args = self._get_declare_var_args(node)

        if node.is_input_stream:
            # Removes the '_input_stream' from the end of the name
            stream_name = node.name[13:]
            self._line(
                f"{node.name} = declare_input_stream({_var_type_dec[node.type]}, '{stream_name}', {_dict_to_python_call(args)})"
            )
        else:
            self._line(f"{node.name} = declare({_var_type_dec[node.type]}, {_dict_to_python_call(args)})")

        return False

    def enter_betterproto_lib_std_google_protobuf_ListValue(
        self, node: betterproto.lib.google.protobuf.ListValue
    ) -> bool:
        line = _stream_processing_terminal_statement(node)
        self._line(line)
        self._fix_legacy_save(line, node)
        return False

    def leave_qm_grpc_qua_QuaProgram(self, node: qm.grpc.qua.QuaProgram) -> bool:
        if self._lines[-1].find("with stream_processing():") > 0:
            self._lines.pop()
        return False

    def _get_declare_var_args(self, node: QuaProgramVarDeclaration) -> Dict[str, str]:
        size = node.size
        dim = node.dim
        args = {}
        if dim > 0:
            args["size"] = f"{size}"
        if dim > 0 and len(node.value) > 0:
            args["value"] = "[" + ", ".join([_ser_exp(it) for it in node.value]) + "]"
        elif len(node.value) == 1:
            args["value"] = _ser_exp(node.value[0])
        if "value" in args and "size" in args:
            del args["size"]
        return args

    def enter_qm_grpc_qua_QuaProgramMeasureStatement(self, node: qua.QuaProgramMeasureStatement) -> bool:
        if node.timestamp_label and node.timestamp_label not in self.tags:
            self._line(f"{node.timestamp_label} = declare_stream()")
            self.tags.append(node.timestamp_label)
        if node.stream_as and node.stream_as not in self.tags:
            if node.stream_as.startswith("atr_"):
                self._line(f"{node.stream_as} = declare_stream(adc_trace=True)")
            else:
                self._line(f"{node.stream_as} = declare_stream()")
            self.tags.append(node.stream_as)
        return self._default_enter(node)

    def visit_qm_grpc_qua_QuaProgramForStatement(self, node: qua.QuaProgramForStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgramForEachStatement(self, node: qua.QuaProgramForEachStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgramIfStatement(self, node: qua.QuaProgramIfStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")
        elseifs_list = node.elseifs
        for elseif in elseifs_list:
            elseif_block = elseif.body
            condition = elseif.condition
            self._leave_block()
            self._line(f"with elif_({ExpressionSerializingVisitor.serialize(condition)}):")
            self._enter_block()
            if len(elseif_block.statements) > 0:
                super()._default_visit(elseif_block)
            else:
                self._line("pass")

        else_block = node.else_
        if len(else_block.statements) > 0:  # Cannot ID else_() with pass
            self._leave_block()
            self._line("with else_():")
            self._enter_block()
            super()._default_visit(else_block)

    def visit_qm_grpc_qua_QuaProgramPlayStatement(self, node: qua.QuaProgramPlayStatement) -> None:
        pulse_one_of, value = betterproto.which_one_of(node, "pulseType")
        if pulse_one_of == "named_pulse":
            pulse = f'"{node.named_pulse.name}"'
            # node.pulse - duplicate with namedPulse
        elif pulse_one_of == "ramp_pulse":
            pulse = f"ramp({ExpressionSerializingVisitor.serialize(node.ramp_pulse)})"
        else:
            raise QmQuaException(f"Unknown pulse type {pulse_one_of}")

        element = node.qe.name
        amp = ""
        if betterproto.serialized_on_wire(node.amp):
            v0 = _ser_exp(node.amp.v0)
            v1 = _ser_exp(node.amp.v1)
            v2 = _ser_exp(node.amp.v2)
            v3 = _ser_exp(node.amp.v3)
            if v0 != "":
                if v1 != "":
                    amp = f"*amp({v0}, {v1}, {v2}, {v3})"
                else:
                    amp = f"*amp({v0})"
        args = []

        _, duration = betterproto.which_one_of(node.duration, "expression_oneof")
        if duration is not None:
            args.append(f"duration={_ser_exp(duration)}")

        _, condition = betterproto.which_one_of(node.condition, "expression_oneof")
        if condition is not None:
            args.append(f"condition={_ser_exp(condition)}")

        if node.target_input:
            args.append(f'target="{node.target_input}"')

        if betterproto.serialized_on_wire(node.chirp):
            rate_one_of, value = betterproto.which_one_of(node.chirp, "rate")
            if isinstance(value, (QuaProgramAnyScalarExpression, QuaProgramArrayVarRefExpression)):
                rate = ExpressionSerializingVisitor.serialize(value)
            else:
                raise QmQuaException(f"Unknown chirp rate {rate_one_of}")

            unit_mapping = {
                qua.QuaProgramChirpUnits.HzPerNanoSec: "Hz/nsec",
                qua.QuaProgramChirpUnits.mHzPerNanoSec: "mHz/nsec",
                qua.QuaProgramChirpUnits.uHzPerNanoSec: "uHz/nsec",
                qua.QuaProgramChirpUnits.nHzPerNanoSec: "nHz/nsec",
                qua.QuaProgramChirpUnits.pHzPerNanoSec: "pHz/nsec",
            }
            s_units = node.chirp.units
            if s_units in unit_mapping:
                units = unit_mapping[s_units]
            else:
                raise QmQuaException(f"Unsupported units {s_units}")

            times_builder = []
            for time in node.chirp.times:
                times_builder.append(f"{time}")

            if len(times_builder) > 0:
                times = f'[{", ".join(times_builder)}]'
            else:
                times = "None"

            if node.chirp.continue_chirp:
                args.append("continue_chirp=True")

            args.append(f'chirp=({rate},{times},"{units}")')

        _, truncate = betterproto.which_one_of(node.truncate, "expression_oneof")
        if truncate is not None:
            args.append(f"truncate={_ser_exp(truncate)}")

        if node.timestamp_label:
            if node.timestamp_label not in self.tags:
                self._line(f"{node.timestamp_label} = declare_stream()")
                self.tags.append(node.timestamp_label)
            args.append(f"timestamp_stream={node.timestamp_label}")

        # TODO maybe make sure no other fields?

        if len(args) > 0:
            args_str = f', {", ".join(args)}'
        else:
            args_str = ""
        indent = ""
        if betterproto.serialized_on_wire(node.port_condition):
            self._line(f"with port_condition({_ser_exp(node.port_condition)}):")
            indent = " " * 4
        self._line(f'{indent}play({pulse}{amp}, "{element}"{args_str})')

    def _default_leave(self, node: Node) -> None:
        if isinstance(node, tuple(_blocks)):
            self._leave_block()
        super()._default_leave(node)

    def _enter_block(self, line: Optional[str] = None) -> None:
        if line is not None:
            self._line(line)
        self._indent += 1

    def _leave_block(self) -> None:
        self._indent -= 1

    def _line(self, line: str) -> None:
        self._lines.append((self._indent * "    ") + line)


def _ser_exp(value: Node) -> str:
    return ExpressionSerializingVisitor.serialize(value)


def _dict_to_python_call(d: Dict[Any, Any]) -> str:
    return ", ".join([f"{k}={v}" for k, v in d.items()])


def _ramp_to_zero_statement(node: QuaProgramRampToZeroStatement) -> str:
    args = []
    args.append(f'"{node.qe.name}"')
    if node.duration is not None:
        args.append(str(node.duration))
    return f'ramp_to_zero({", ".join(args)})'


def _measure_statement(node: qua.QuaProgramMeasureStatement) -> str:
    args = []

    amp = ""
    v0 = _ser_exp(node.amp.v0)
    v1 = _ser_exp(node.amp.v1)
    v2 = _ser_exp(node.amp.v2)
    v3 = _ser_exp(node.amp.v3)
    if v0 != "":
        if v1 != "":
            amp = f"*amp({v0}, {v1}, {v2}, {v3})"
        else:
            amp = f"*amp({v0})"

    args.append(f'"{node.pulse.name}"{amp}')
    args.append(f'"{node.qe.name}"')
    if node.stream_as:
        args.append(f"{node.stream_as}")
    else:
        args.append("None")

    if len(node.measure_processes) > 0:
        for process in node.measure_processes:
            args.append(ExpressionSerializingVisitor.serialize(process))
    if node.timestamp_label:
        args.append(f"timestamp_stream={node.timestamp_label}")
    return f'measure({", ".join(args)})'


def _wait_statement(node: qua.QuaProgramWaitStatement) -> str:
    args = []
    _, wait_value = betterproto.which_one_of(node.time, "expression_oneof")
    if wait_value is not None:
        args.append(f"{ExpressionSerializingVisitor.serialize(wait_value)}")
    qes = []
    for qe in node.qe:
        qes.append(f'"{qe.name}"')
    args.append(", ".join(qes))
    return f'wait({", ".join(args)})'


def _align_statement(node: QuaProgramAlignStatement) -> str:
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    return f'align({", ".join(args)})'


def _wait_for_trigger_statement(node: qua.QuaProgramWaitForTriggerStatement) -> str:
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    if node.pulse_to_play.name:
        args.append(f'"{node.pulse_to_play.name}"')
    output = betterproto.which_one_of(node, "source")[1]
    if isinstance(output, QuaProgramWaitForTriggerStatementElementOutput) and output.element:
        if node.element_output.output:
            args.append(f'trigger_element=("{node.element_output.element}", "{node.element_output.output}")')
        else:
            args.append(f'trigger_element="{node.element_output.element}"')
    if betterproto.which_one_of(node.time_tag_target, "var_oneof")[0] == "name":
        args.append(f"time_tag_target={node.time_tag_target.name}")
    return f'wait_for_trigger({", ".join(args)})'


def _frame_rotation_statement(node: QuaProgramZRotationStatement) -> str:
    args = []
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    args.append(f'"{node.qe.name}"')
    return f'frame_rotation_2pi({", ".join(args)})'


def _fast_frame_rotation_statement(node: QuaProgramFastFrameRotationStatement) -> str:
    args = []
    args.append(f"{ExpressionSerializingVisitor.serialize(node.cosine)}")
    args.append(f"{ExpressionSerializingVisitor.serialize(node.sine)}")
    args.append(f'"{node.qe.name}"')
    return f'fast_frame_rotation({", ".join(args)})'


def _reset_frame_statement(node: QuaProgramResetFrameStatement) -> str:
    args = []
    args.append(f'"{node.qe.name}"')
    return f'reset_frame({", ".join(args)})'


def _update_frequency_statement(node: QuaProgramUpdateFrequencyStatement) -> str:
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    args.append(f'"{qua.QuaProgramUpdateFrequencyStatementUnits(node.units).name}"')
    args.append(f"{node.keep_phase}")
    return f'update_frequency({", ".join(args)})'


def _set_dc_offset_statement(node: qua.QuaProgramSetDcOffsetStatement) -> str:
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f'"{node.qe_input_reference}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.offset)}")
    return f'set_dc_offset({", ".join(args)})'


def _advance_input_stream_statement(node: QuaProgramAdvanceInputStreamStatement) -> str:
    stream_value = betterproto.which_one_of(node, "stream_oneof")[1]
    if isinstance(stream_value, QuaProgramArrayVarRefExpression) and stream_value.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(stream_value)
    elif isinstance(stream_value, QuaProgramVarRefExpression) and stream_value.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(stream_value)
    else:
        raise RuntimeError("unsupported type for pop input stream")
    return f"advance_input_stream({input_stream})"


def _for_block_statement(node: QuaProgramForStatement) -> str:
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if len(node.init.statements) == 0 and len(node.update.statements) == 0:
        if condition == "True":
            return "with infinite_loop_():"
        else:
            return f"with while_({condition}):"
    else:
        if len(node.init.statements) != 1:
            raise Exception("for is not valid")
        if len(node.update.statements) != 1:
            raise Exception("for is not valid")

        init = node.init.statements[0].assign
        update = node.update.statements[0].assign

        if init is None:
            raise Exception("for is not valid")
        if update is None:
            raise Exception("for is not valid")

        return f"with for_({ExpressionSerializingVisitor.serialize(init.target)},{ExpressionSerializingVisitor.serialize(init.expression)},{condition},{ExpressionSerializingVisitor.serialize(update.expression)}):"


def _for_each_block_statement(node: QuaProgramForEachStatement) -> str:
    variables = []
    arrays = []
    for it in node.iterator:
        variables.append(_ser_exp(it.variable))
        arrays.append(it.array.name)
    return f'with for_each_(({",".join(variables)}),({",".join(arrays)})):'


def _if_block_statement(node: QuaProgramIfStatement) -> str:
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if node.unsafe is True:
        unsafe = ", unsafe=True"
    else:
        unsafe = ""
    return f"with if_({condition}{unsafe}):"


def _strict_timing_block_statement(_: Any) -> str:
    return "with strict_timing_():"


def _stream_processing_function(array: List[Value]) -> str:
    function = array[0].string_value

    if function == "average":
        if len(array) > 1:
            first_element = array[1]
            v = betterproto.which_one_of(first_element, "kind")[1]
            if isinstance(v, str):
                var = v
            elif isinstance(v, ListValue):
                var = _stream_processing_operator(v.values)
            else:
                raise NotImplementedError
        else:
            var = ""
        return f"average({var})"

    if function == "dot":
        if len(array) == 1:
            return "tuple_dot_product()"
        else:
            vector = _stream_processing_operator(array[1].list_value.values)
            return f"dot_product({vector})"

    if function == "vmult":
        vector = _stream_processing_operator(array[1].list_value.values)
        return f"multiply_by({vector})"

    if function == "smult":
        return f"multiply_by({array[1].string_value})"

    if function == "tmult":
        return "tuple_multiply()"

    if function == "conv":
        if len(array) == 2:
            if array[1].string_value:
                mode = f'"{array[1].string_value}"'
            else:
                mode = ""
            return f"tuple_convolution({mode})"
        else:
            vector = _stream_processing_operator(array[2].list_value.values)
            if array[1].string_value:
                mode = f',"{array[1].string_value}"'
            else:
                mode = ""
            return f"convolution({vector}{mode})"

    if function == "fft":
        return "fft()"

    if function == "booleancast":
        return "boolean_to_int()"

    if function == "demod":
        if len(array[2].list_value.values) > 0:
            cos = _stream_processing_operator(array[2].list_value.values)
        else:
            cos = array[2].string_value
        if len(array[3].list_value.values) > 0:
            sin = _stream_processing_operator(array[3].list_value.values)
        else:
            sin = array[3].string_value
        return f"demod({array[1].string_value},{cos},{sin})"

    print(f"missing function: {function}")
    return "default_function()"


def _stream_processing_operator(array: List[Value]) -> str:
    operator = array[0].string_value

    if operator == "@macro_input":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.input{array[1].string_value}()"

    if operator == "@array":
        values = []
        for a in array[1:]:
            value = _stream_processing_statement(a)
            values.append(value)
        return f'[{", ".join(values)}]'

    if operator == "@macro_auto_reshape":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.auto_reshape()"

    if operator == "+":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.add({right})"  # arithmetic stream

    if operator == "-":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.subtract({right})"  # arithmetic stream

    if operator == "/":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.divide({right})"  # arithmetic stream

    if operator == "*":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.multiply({right})"  # arithmetic stream

    if operator == "zip":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{right}.zip({left})"  # arithmetic stream

    if operator == "take":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.take({array[1].string_value})"

    if operator == "buffer":
        chain = _default_stream_processing_chain(array)
        dims = [dim.string_value for dim in array[1:-1]]
        return f'{chain}.buffer({", ".join(dims)})'

    if operator == "bufferAndSkip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.buffer_and_skip({array[1].string_value}, {array[2].string_value})"

    if operator == "skip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip({array[1].string_value})"

    if operator == "skipLast":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip_last({array[1].string_value})"

    if operator == "histogram":
        chain = _default_stream_processing_chain(array)
        bins = _stream_processing_operator(array[1].list_value.values)
        return f"{chain}.histogram({bins})"

    if operator == "average":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.average()"

    if operator == "real":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.real()"

    if operator == "image":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.image()"

    if operator == "flatten":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.flatten()"

    if operator == "map":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.map(FUNCTIONS.{_stream_processing_function(array[1].list_value.values)})"

    if operator == "@re":
        chain = _default_stream_processing_chain(array)
        timestamp_mode = int(array[1].string_value)
        if timestamp_mode == 0:  # values
            return f"{chain}"
        elif timestamp_mode == 1:  # timestamps
            return f"{chain}.timestamps()"
        elif timestamp_mode == 2:  # ValuesAndTimestamps
            return f"{chain}.with_timestamps()"

    if operator == "@macro_adc_trace":
        chain = _default_stream_processing_chain(array)
        return f"{chain}"

    print(f"missing operator: {operator}")
    chain = _default_stream_processing_chain(array)
    return f"{chain}"


def _default_stream_processing_chain(array: List[Value]) -> str:
    last_index = len(array) - 1
    chain = _stream_processing_statement(array[last_index])
    return f"{chain}"


def _stream_processing_statement(node: Value) -> str:
    _value = betterproto.which_one_of(node, "kind")[1]
    if isinstance(_value, ListValue) and len(_value.values) > 0:
        return _stream_processing_operator(_value.values)
    elif isinstance(_value, str):
        return _value
    else:
        raise NotImplementedError(f"Unsupported stream processing statement: {_value}")


def _stream_processing_terminal_statement(node: ListValue) -> str:
    last_index = len(node.values) - 1
    chain = _stream_processing_statement(node.values[last_index])
    terminal = node.values[0].string_value
    terminal = "save_all" if terminal == "saveAll" else terminal  # normalize save all
    return f'{chain}.{terminal}("{node.values[1].string_value}")'


def _update_correction_statement(node: QuaProgramUpdateCorrectionStatement) -> str:
    return (
        f'update_correction("{node.qe.name}",{ExpressionSerializingVisitor.serialize(node.correction.c0)},'
        f"{ExpressionSerializingVisitor.serialize(node.correction.c1)},"
        f"{ExpressionSerializingVisitor.serialize(node.correction.c2)},"
        f"{ExpressionSerializingVisitor.serialize(node.correction.c3)})"
    )


def _assignment_statement(node: QuaProgramAssignmentStatement) -> str:
    return (
        f"assign({ExpressionSerializingVisitor.serialize(node.target)}, "
        f"{ExpressionSerializingVisitor.serialize(node.expression)})"
    )


def _serialize(node: Node) -> List[str]:
    visitor = QuaSerializingVisitor()
    visitor.visit(node)
    return visitor._out_lines()


_blocks = {
    QuaProgram: lambda n: "with program() as prog:",
    QuaProgramForStatement: _for_block_statement,
    QuaProgramForEachStatement: _for_each_block_statement,
    QuaProgramIfStatement: _if_block_statement,
    QuaProgramStrictTimingStatement: _strict_timing_block_statement,
}


_statements = {
    QuaProgramMeasureStatement: _measure_statement,
    QuaProgramWaitStatement: _wait_statement,
    QuaProgramAssignmentStatement: _assignment_statement,
    QuaProgramPauseStatement: lambda n: "pause()",
    QuaProgramResetPhaseStatement: lambda n: f'reset_if_phase("{n.qe.name}")',
    QuaProgramResetGlobalPhaseStatement: lambda n: "reset_global_phase()",
    QuaProgramUpdateFrequencyStatement: _update_frequency_statement,
    QuaProgramAlignStatement: _align_statement,
    QuaProgramWaitForTriggerStatement: _wait_for_trigger_statement,
    QuaProgramZRotationStatement: _frame_rotation_statement,
    QuaProgramRampToZeroStatement: _ramp_to_zero_statement,
    QuaProgramResetFrameStatement: _reset_frame_statement,
    ListValue: _stream_processing_terminal_statement,
    # list value statement is assumed just as stream processing for now
    QuaProgramUpdateCorrectionStatement: _update_correction_statement,
    QuaProgramSetDcOffsetStatement: _set_dc_offset_statement,
    QuaProgramAdvanceInputStreamStatement: _advance_input_stream_statement,
    QuaProgramFastFrameRotationStatement: _fast_frame_rotation_statement,
}

_var_type_dec = {QuaProgramType.INT: "int", QuaProgramType.BOOL: "bool", QuaProgramType.REAL: "fixed"}

_nodes_to_ignore = {
    QuaProgramScript,
    QuaProgramStatementsCollection,
    QuaProgramAnyStatement,
    QuaProgramAnyScalarExpression,
    QuaProgramBinaryExpression,
    QuaProgramVarRefExpression,
    QuaProgramLiteralExpression,
    QuaProgramPlayStatement,
    QuaConfig,
}

_dont_print = set(_blocks) | set(_statements) | _nodes_to_ignore



import warnings
import json as _json
from io import BytesIO
from typing import Any, Dict, Union, BinaryIO, Callable, Optional

import numpy
import betterproto
import numpy.typing
from numpy.lib import format as _format
from betterproto.lib.google.protobuf import Value, Struct

from qm.persistence import BaseStore
from qm.utils import deprecation_message
from qm.utils.async_utils import run_async
from qm.api.frontend_api import FrontendApi
from qm.exceptions import QMSimulationError
from qm.waveform_report import WaveformReport
from qm.api.simulation_api import SimulationApi
from qm.jobs.running_qm_job import RunningQmJob
from qm.grpc.frontend import SimulatedResponsePart
from qm.api.models.capabilities import ServerCapabilities
from qm.results.simulator_samples import SimulatorSamples
from qm.grpc.v2 import PullSamplesResponsePullSamplesResponseSuccess
from qm.type_hinting.simulator_types import AnalogOutputsType, DigitalOutputsType, WaveformInPortsType
from qm.grpc.results_analyser import (
    SimulatorSamplesResponse,
    SimulatorSamplesResponseData,
    SimulatorSamplesResponseHeader,
)


class SimulatorOutput:
    def __init__(self, job_id: str, frontend: FrontendApi) -> None:
        super().__init__()
        self._id = job_id
        self._simulation_api = SimulationApi(frontend.connection_details)


def extract_value(value: Value) -> Any:
    name, one_of = betterproto.which_one_of(value, "kind")
    if name in VALUE_MAPPING:
        return VALUE_MAPPING[name](one_of)


def extract_struct_value(struct_value: Struct) -> Any:
    output = {}
    for name, value in struct_value.fields.items():
        output[name] = extract_value(value)
    return output


VALUE_MAPPING: Dict[str, Callable[[Any], Any]] = {
    "number_value": float,
    "string_value": str,
    "bool_value": bool,
    "list_value": lambda list_value: [extract_value(value) for value in list_value.values],
    "struct_value": extract_struct_value,
    "null_value": lambda x: None,
}


class SimulatedJob(RunningQmJob):
    def __init__(
        self,
        job_id: str,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
        simulated_response: SimulatedResponsePart,
    ):
        super().__init__(job_id, "", frontend_api, capabilities, store)
        self._waveform_report: Optional[WaveformReport] = None

        self._simulated_analog_outputs: AnalogOutputsType = {"waveforms": None}
        self._simulated_digital_outputs: DigitalOutputsType = {"waveforms": None}
        if betterproto.serialized_on_wire(simulated_response.analog_outputs):
            self._simulated_analog_outputs = extract_struct_value(simulated_response.analog_outputs)
        if betterproto.serialized_on_wire(simulated_response.digital_outputs):
            self._simulated_digital_outputs = extract_struct_value(simulated_response.digital_outputs)
        if betterproto.serialized_on_wire(simulated_response.waveform_report):
            self._waveform_report = WaveformReport.from_dict(
                extract_struct_value(simulated_response.waveform_report), self.id
            )
        if simulated_response.errors:
            raise QMSimulationError("\n".join(simulated_response.errors))

        self._simulation_api = SimulationApi(self._frontend.connection_details)

    def _initialize_from_job_status(self) -> None:
        # Overriding this for simulated jobs to do nothing
        pass

    def get_simulated_waveform_report(self) -> Optional[WaveformReport]:
        """
        Get this Job's Waveform report.
        Returns:
            The waveform report. `None` will be returned in case of an error.
        """
        return self._waveform_report

    def simulated_analog_waveforms(self) -> Optional[WaveformInPortsType]:
        """
        Return the results of the simulation of elements and analog outputs.

        The returned dictionary has the following keys and entries:

        - **elements**: a dictionary containing the outputs with timestamps and values arranged by elements.

        - **controllers: a dictionary containing the outputs with timestamps and values arranged by controllers.

            - **ports**: a dictionary containing the outputs with timestamps and values arranged by output ports.
                for each element or output port, the entry is a list of dictionaries with the following information:

                - **timestamp**:
                    The time, in nsec, from the start of the program to the start of the pulse.

                - **samples**:

                    Output information, with ``duration`` given in nsec and ``value`` given normalized OPX output units.

        Returns:
            A dictionary containing output information for the analog outputs of the controller.

        """
        warnings.warn(
            deprecation_message(
                method="SimulatedJob.simulated_analog_waveforms",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use 'get_simulated_waveform_report' instead.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self._simulated_analog_outputs["waveforms"]

    def simulated_digital_waveforms(self) -> Optional[WaveformInPortsType]:
        """
        Return the results of the simulation of digital outputs.

        - **controllers**: a dictionary containing the outputs with timestamps and values arranged by controllers.

            - **ports**: a dictionary containing the outputs with timestamps and values arranged by output ports.
                for each element or output port, the entry is a list of dictionaries with the following information:

                - **timestamp**:
                    The time, in nsec, from the start of the program to the start of the pulse.

                - **samples**:
                    A list containing the sequence of outputted values, with ``duration`` given in nsec
                    and ``value`` given as a boolean value.

        Returns:
            A dictionary containing output information for the analog outputs of the controller.
        """
        warnings.warn(
            deprecation_message(
                method="SimulatedJob.simulated_digital_waveforms",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use 'get_simulated_waveform_report' instead.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return self._simulated_digital_outputs["waveforms"]

    async def _pull_simulator_samples(
        self, include_analog: bool, include_digital: bool, writer: BinaryIO, data_writer: BinaryIO
    ) -> None:
        async for result in self._simulation_api.pull_simulator_samples(self._id, include_analog, include_digital):
            if result.ok:
                _write_simulator_result(result, writer, data_writer)
            else:
                raise QMSimulationError("Error while pulling samples")

    def _get_np_simulated_samples(
        self, include_analog: bool = True, include_digital: bool = True
    ) -> numpy.typing.NDArray[numpy.generic]:
        writer = BytesIO()
        data_writer = BytesIO()

        run_async(self._pull_simulator_samples(include_analog, include_digital, writer, data_writer))
        return _build_np_array(writer, data_writer)

    def get_simulated_samples(self, include_analog: bool = True, include_digital: bool = True) -> SimulatorSamples:
        """
        Obtain the output samples of a QUA program simulation.

        Samples are returned in an object that holds the controllers in the current simulation,
        where each controller's name will be a property of this object.
        The value of each property of the returned value is an object with the following properties:

        ``analog``:

            holds a dictionary with analog port names as keys and numpy array of samples as values.

        ``digital``:

            holds a dictionary with digital port names as keys and numpy array of samples as values.

        It is also possible to directly plot the outputs using a built-in plot command.

        Example:
            ```python
            samples = job.get_simulated_samples()
            analog1 = samples.con1.analog["1"]  # obtain analog port 1 of controller "con1"
            digital9 = samples.con1.analog["9"]  # obtain digital port 9 of controller "con1"
            samples.con1.plot()  # Plots all active ports
            samples.con1.plot(analog_ports=['1', '2'], digital_ports=['9'])  # Plots the given output ports
            ```

        !!! Note

            The simulated digital waveforms are delayed by 136ns relative to the real
            output of the OPX.

        Args:
            include_analog: Should we collect simulated analog samples
            include_digital: Should we collect simulated digital samples

        Returns:
            The simulated samples of the job.
        """
        return SimulatorSamples.from_np_array(
            self._get_np_simulated_samples(include_analog=include_analog, include_digital=include_digital),
        )

    @property
    def simulator(self) -> SimulatorOutput:
        return SimulatorOutput(self._id, self._frontend)

    def plot_waveform_report_with_simulated_samples(self) -> None:
        if self._waveform_report is None:
            raise QMSimulationError("Waveform report is not available")
        self._waveform_report.create_plot(self.get_simulated_samples())

    def plot_waveform_report_without_samples(self) -> None:
        if self._waveform_report is None:
            raise QMSimulationError("Waveform report is not available")
        self._waveform_report.create_plot()


def _write_simulator_result(
    result: Union[SimulatorSamplesResponse, PullSamplesResponsePullSamplesResponseSuccess],
    writer: BinaryIO,
    data_writer: BinaryIO,
) -> None:
    _, value = betterproto.which_one_of(result, "body")
    if isinstance(value, SimulatorSamplesResponseHeader):
        _format.write_array_header_2_0(  # type: ignore[no-untyped-call]
            writer,
            {
                "descr": _json.loads(value.simple_d_type),
                "fortran_order": False,
                "shape": (value.count_of_items,),
            },
        )
    elif isinstance(value, SimulatorSamplesResponseData):
        data_writer.write(value.data)
    else:
        raise QMSimulationError("Error while pulling samples")


def _build_np_array(writer: BinaryIO, data_writer: BinaryIO) -> numpy.typing.NDArray[numpy.generic]:
    data_writer.seek(0)
    for d in data_writer:
        writer.write(d)

    writer.seek(0)
    ret: numpy.typing.NDArray[numpy.generic] = numpy.load(writer)
    return ret



from typing import List, Tuple, Union, Literal, Mapping, TypedDict

from qm.type_hinting.general import Number

StandardPort = Tuple[str, int, int]
PortReferenceType = Union[Tuple[str, int], StandardPort]


# TODO: This is a placeholder while we still use dicts, once we move to pydantics we can simply change the
#  inheritance of the classes handled here and add a more robust validation to the types


class AnalogOutputFilterConfigType(TypedDict, total=False):
    feedforward: List[float]
    feedback: List[float]


class AnalogOutputPortConfigType(TypedDict, total=False):
    offset: Number
    filter: AnalogOutputFilterConfigType
    delay: int
    crosstalk: Mapping[int, Number]
    shareable: bool


class AnalogInputPortConfigType(TypedDict, total=False):
    offset: Number
    gain_db: int
    shareable: bool
    sampling_rate: float


class DigitalOutputPortConfigType(TypedDict, total=False):
    shareable: bool
    inverted: bool
    level: Literal["TTL", "LVTTL"]


class DigitalInputPortConfigType(TypedDict, total=False):
    shareable: bool
    deadtime: int
    polarity: Literal["RISING", "FALLING"]
    threshold: Number


class AnalogOutputPortConfigTypeOctoDac(TypedDict, total=False):
    offset: Number
    filter: AnalogOutputFilterConfigType
    delay: int
    crosstalk: Mapping[int, Number]
    shareable: bool
    connectivity: Tuple[str, str]
    sampling_rate: float
    upsampling_mode: Literal["mw", "pulse"]
    output_mode: Literal["direct", "amplified"]


class LfFemConfigType(TypedDict, total=False):
    type: Literal["LF"]
    analog_outputs: Mapping[Union[int, str], AnalogOutputPortConfigTypeOctoDac]
    analog_inputs: Mapping[Union[int, str], AnalogInputPortConfigType]
    digital_outputs: Mapping[Union[int, str], DigitalOutputPortConfigType]
    digital_inputs: Mapping[Union[int, str], DigitalInputPortConfigType]


Band = Literal[1, 2, 3]


class MwFemAnalogInputPortConfigType(TypedDict, total=False):
    sampling_rate: float
    gain_db: int
    shareable: bool
    band: Band
    downconverter_frequency: float


class MwUpconverterConfigType(TypedDict, total=False):
    frequency: float


class MwFemAnalogOutputPortConfigType(TypedDict, total=False):
    sampling_rate: float
    full_scale_power_dbm: int
    band: Band
    delay: int
    shareable: bool
    upconverters: Mapping[int, MwUpconverterConfigType]
    upconverter_frequency: float


class MwFemConfigType(TypedDict, total=False):
    type: Literal["MW"]
    analog_outputs: Mapping[Union[int, str], MwFemAnalogOutputPortConfigType]
    analog_inputs: Mapping[Union[int, str], MwFemAnalogInputPortConfigType]
    digital_outputs: Mapping[Union[int, str], DigitalOutputPortConfigType]
    digital_inputs: Mapping[Union[int, str], DigitalInputPortConfigType]


class ControllerConfigType(TypedDict, total=False):
    type: Literal["opx", "opx1"]
    analog_outputs: Mapping[Union[int, str], AnalogOutputPortConfigType]
    analog_inputs: Mapping[Union[int, str], AnalogInputPortConfigType]
    digital_outputs: Mapping[Union[int, str], DigitalOutputPortConfigType]
    digital_inputs: Mapping[Union[int, str], DigitalInputPortConfigType]


class OctaveRFOutputConfigType(TypedDict, total=False):
    LO_frequency: float
    LO_source: Literal["internal", "external"]
    output_mode: Literal["always_on", "always_off", "triggered", "triggered_reversed"]
    gain: Union[int, float]
    input_attenuators: Literal["ON", "OFF"]
    I_connection: PortReferenceType
    Q_connection: PortReferenceType


class OctaveRFInputConfigType(TypedDict, total=False):
    RF_source: Literal["RF_in", "loopback_1", "loopback_2", "loopback_3", "loopback_4", "loopback_5"]
    LO_frequency: float
    LO_source: Literal["internal", "external", "analyzer"]
    IF_mode_I: Literal["direct", "mixer", "envelope", "off"]
    IF_mode_Q: Literal["direct", "mixer", "envelope", "off"]


class OctaveSingleIfOutputConfigType(TypedDict, total=False):
    port: PortReferenceType
    name: str


class OctaveIfOutputsConfigType(TypedDict, total=False):
    IF_out1: OctaveSingleIfOutputConfigType
    IF_out2: OctaveSingleIfOutputConfigType


FEM_IDX = Literal[1, 2, 3, 4, 5, 6, 7, 8, "1", "2", "3", "4", "5", "6", "7", "8"]


class OPX1000ControllerConfigType(TypedDict, total=False):
    type: Literal["opx1000"]
    fems: Mapping[FEM_IDX, Union[LfFemConfigType, MwFemConfigType]]


LoopbackType = Tuple[
    Tuple[str, Literal["Synth1", "Synth2", "Synth3", "Synth4", "Synth5"]],
    Literal["Dmd1LO", "Dmd2LO", "LO1", "LO2", "LO3", "LO4", "LO5"],
]


class OctaveConfigType(TypedDict, total=False):
    RF_outputs: Mapping[int, OctaveRFOutputConfigType]
    RF_inputs: Mapping[int, OctaveRFInputConfigType]
    IF_outputs: OctaveIfOutputsConfigType
    loopbacks: List[LoopbackType]
    connectivity: Union[str, Tuple[str, int]]


class DigitalInputConfigType(TypedDict, total=False):
    delay: int
    buffer: int
    port: PortReferenceType


class IntegrationWeightConfigType(TypedDict, total=False):
    cosine: Union[List[Tuple[float, int]], List[float]]
    sine: Union[List[Tuple[float, int]], List[float]]


class ConstantWaveFormConfigType(TypedDict, total=False):
    type: Literal["constant"]
    sample: float


class CompressedWaveFormConfigType(TypedDict, total=False):
    type: str
    samples: List[float]
    sample_rate: float


class ArbitraryWaveFormConfigType(TypedDict, total=False):
    type: Literal["arbitrary"]
    samples: List[float]
    max_allowed_error: float
    sampling_rate: Number
    is_overridable: bool


class DigitalWaveformConfigType(TypedDict, total=False):
    samples: List[Tuple[int, int]]


class MixerConfigType(TypedDict, total=False):
    intermediate_frequency: float
    lo_frequency: float
    correction: Tuple[Number, Number, Number, Number]


class PulseConfigType(TypedDict, total=False):
    operation: Literal["measurement", "control"]
    length: int
    waveforms: Mapping[str, str]
    digital_marker: str
    integration_weights: Mapping[str, str]


class SingleInputConfigType(TypedDict, total=False):
    port: PortReferenceType


class MwInputConfigType(TypedDict, total=False):
    port: PortReferenceType
    upconverter: int


class MwOutputConfigType(TypedDict, total=False):
    port: PortReferenceType


class HoldOffsetConfigType(TypedDict, total=False):
    duration: int


class StickyConfigType(TypedDict, total=False):
    analog: bool
    digital: bool
    duration: int


class MixInputConfigType(TypedDict, total=False):
    I: PortReferenceType
    Q: PortReferenceType
    mixer: str
    lo_frequency: float


class InputCollectionConfigType(TypedDict, total=False):
    inputs: Mapping[str, PortReferenceType]


class OscillatorConfigType(TypedDict, total=False):
    intermediate_frequency: float
    mixer: str
    lo_frequency: float


class OutputPulseParameterConfigType(TypedDict):
    signalThreshold: int
    signalPolarity: Literal["ABOVE", "ASCENDING", "BELOW", "DESCENDING"]
    derivativeThreshold: int
    derivativePolarity: Literal["ABOVE", "ASCENDING", "BELOW", "DESCENDING"]


class ElementConfigType(TypedDict, total=False):
    intermediate_frequency: float
    oscillator: str
    measurement_qe: str
    operations: Mapping[str, str]
    singleInput: SingleInputConfigType
    mixInputs: MixInputConfigType
    singleInputCollection: InputCollectionConfigType
    multipleInputs: InputCollectionConfigType
    MWInput: MwInputConfigType
    MWOutput: MwOutputConfigType
    time_of_flight: int
    smearing: int
    outputs: Mapping[str, PortReferenceType]
    digitalInputs: Mapping[str, DigitalInputConfigType]
    digitalOutputs: Mapping[str, PortReferenceType]
    outputPulseParameters: OutputPulseParameterConfigType
    hold_offset: HoldOffsetConfigType
    sticky: StickyConfigType
    thread: str
    RF_inputs: Mapping[str, Tuple[str, int]]
    RF_outputs: Mapping[str, Tuple[str, int]]


class DictQuaConfig(TypedDict, total=False):
    version: Literal[1, "1"]
    oscillators: Mapping[str, OscillatorConfigType]
    elements: Mapping[str, ElementConfigType]
    controllers: Mapping[str, Union[ControllerConfigType, OPX1000ControllerConfigType]]
    octaves: Mapping[str, OctaveConfigType]
    integration_weights: Mapping[str, IntegrationWeightConfigType]
    waveforms: Mapping[
        str, Union[ArbitraryWaveFormConfigType, ConstantWaveFormConfigType, CompressedWaveFormConfigType]
    ]
    digital_waveforms: Mapping[str, DigitalWaveformConfigType]
    pulses: Mapping[str, PulseConfigType]
    mixers: Mapping[str, List[MixerConfigType]]



import dataclasses
from pathlib import Path
from typing import List, Union, Optional

from qm import DictQuaConfig
from qm.grpc.qua_config import QuaConfig
from qm.program._qua_config_schema import load_config
from qm.program._ResultAnalysis import _ResultAnalysis
from qm.program.StatementsCollection import StatementsCollection
from qm.grpc.qua import (
    QuaProgram,
    QuaProgramType,
    QuaProgramScript,
    QuaResultAnalysis,
    QuaProgramVarDeclaration,
    QuaProgramLiteralExpression,
    QuaProgramStatementsCollection,
)


@dataclasses.dataclass
class ProgramMetadata:
    uses_command_timestamps: bool
    uses_fast_frame_rotation: bool


class Program:
    def __init__(
        self,
        config: Optional[QuaConfig] = None,
        program: Optional[QuaProgram] = None,
    ):
        if program is None:
            program = QuaProgram(
                script=QuaProgramScript(variables=[], body=QuaProgramStatementsCollection(statements=[])),
                result_analysis=QuaResultAnalysis(model=[]),
            )

        self._program = program
        self._qua_config = config
        self._result_analysis = _ResultAnalysis(self._program.result_analysis)
        self._is_in_scope = False
        self._metadata = ProgramMetadata(uses_command_timestamps=False, uses_fast_frame_rotation=False)

    def _declare_var(
        self,
        name: str,
        var_type: QuaProgramType,
        size: int,
        value: Union[QuaProgramLiteralExpression, List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ) -> None:
        if not isinstance(value, list):
            value = [value]
        declaration = QuaProgramVarDeclaration(
            name=name,
            type=var_type,
            size=size,
            dim=dim,
            is_input_stream=is_input_stream,
            value=list(value),
        )
        self._program.script.variables.append(declaration)

    def add_declaration(self, declaration: QuaProgramVarDeclaration) -> None:
        self._program.script.variables.append(declaration)

    def declare_int(
        self,
        name: str,
        size: int,
        value: Union[QuaProgramLiteralExpression, List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ) -> None:
        self._declare_var(name, QuaProgramType.INT, size, value, dim, is_input_stream)

    def declare_real(
        self,
        name: str,
        size: int,
        value: Union[QuaProgramLiteralExpression, List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ) -> None:
        self._declare_var(name, QuaProgramType.REAL, size, value, dim, is_input_stream)

    def declare_bool(
        self,
        name: str,
        size: int,
        value: Union[QuaProgramLiteralExpression, List[QuaProgramLiteralExpression]],
        dim: int,
        is_input_stream: bool,
    ) -> None:
        self._declare_var(name, QuaProgramType.BOOL, size, value, dim, is_input_stream)

    @property
    def body(self) -> StatementsCollection:
        return StatementsCollection(self._program.script.body)

    @property
    def result_analysis(self) -> _ResultAnalysis:
        return self._result_analysis

    @property
    def qua_program(self) -> QuaProgram:
        return self._program

    def to_protobuf(self, config: DictQuaConfig) -> bytes:
        """
        Serialize the program to a protobuf binary.
        """
        loaded_config = load_config(config)
        copy = QuaProgram().from_dict(self._program.to_dict())
        copy.config = QuaConfig().from_dict(loaded_config.to_dict())
        return bytes(copy)

    @classmethod
    def from_protobuf(cls, binary: bytes) -> "Program":
        """
        Deserialize the program from a protobuf binary.
        """
        program = QuaProgram().parse(binary)
        return cls(program=program, config=program.config)

    def to_file(self, path: Union[str, Path], config: DictQuaConfig) -> None:
        """
        Serialize the program to a protobuf binary and write it to a file.
        """
        if isinstance(path, str):
            path = Path(path)
        path.write_bytes(self.to_protobuf(config))

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "Program":
        """
        Deserialize the program from a protobuf binary file.
        """
        if isinstance(path, str):
            path = Path(path)
        return cls.from_protobuf(path.read_bytes())

    def set_in_scope(self) -> None:
        self._is_in_scope = True

    def set_exit_scope(self) -> None:
        self._is_in_scope = False

    def is_in_scope(self) -> bool:
        return self._is_in_scope

    @property
    def metadata(self) -> ProgramMetadata:
        return self._metadata

    def set_metadata(
        self,
        uses_command_timestamps: Optional[bool] = None,
        uses_fast_frame_rotation: Optional[bool] = None,
    ) -> None:
        if uses_command_timestamps is not None:
            self._metadata.uses_command_timestamps = uses_command_timestamps
        if uses_fast_frame_rotation is not None:
            self.metadata.uses_fast_frame_rotation = uses_fast_frame_rotation



import uuid
import numbers
from typing import Dict, List, Type, Tuple, Union, TypeVar, Optional, cast

import betterproto
from betterproto.lib.google.protobuf import Empty
from dependency_injector.wiring import Provide, inject

from qm.utils.config_utils import get_fem_config_instance
from qm.utils.list_compression_utils import split_list_to_chunks
from qm.api.models.capabilities import OPX_FEM_IDX, ServerCapabilities
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.exceptions import (
    InvalidOctaveParameter,
    NoInputsOrOutputsError,
    ConfigValidationException,
    OctaveConnectionAmbiguity,
    ElementInputConnectionAmbiguity,
    ElementOutputConnectionAmbiguity,
)
from qm.program._validate_config_schema import (
    validate_oscillator,
    validate_output_tof,
    validate_used_inputs,
    validate_output_smearing,
    validate_sticky_duration,
    validate_arbitrary_waveform,
    validate_timetagging_parameters,
)
from qm.type_hinting.config_types import (
    LoopbackType,
    StandardPort,
    DictQuaConfig,
    LfFemConfigType,
    MixerConfigType,
    MwFemConfigType,
    PulseConfigType,
    OctaveConfigType,
    ElementConfigType,
    PortReferenceType,
    ControllerConfigType,
    OscillatorConfigType,
    DigitalInputConfigType,
    MwUpconverterConfigType,
    OctaveRFInputConfigType,
    OctaveRFOutputConfigType,
    AnalogInputPortConfigType,
    DigitalWaveformConfigType,
    OctaveIfOutputsConfigType,
    AnalogOutputPortConfigType,
    ConstantWaveFormConfigType,
    DigitalInputPortConfigType,
    ArbitraryWaveFormConfigType,
    DigitalOutputPortConfigType,
    IntegrationWeightConfigType,
    OPX1000ControllerConfigType,
    MwFemAnalogInputPortConfigType,
    OctaveSingleIfOutputConfigType,
    MwFemAnalogOutputPortConfigType,
    AnalogOutputPortConfigTypeOctoDac,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMatrix,
    QuaConfigSticky,
    QuaConfigFemTypes,
    QuaConfigMixerDec,
    QuaConfigMixerRef,
    QuaConfigPulseDec,
    QuaConfigDeviceDec,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigHoldOffset,
    QuaConfigOscillator,
    QuaConfigQuaConfigV1,
    QuaConfigSingleInput,
    QuaConfigWaveformDec,
    QuaConfigOctaveConfig,
    QuaConfigOctaveIfMode,
    QuaConfigVoltageLevel,
    QuaConfigControllerDec,
    QuaConfigElementThread,
    QuaConfigOctoDacFemDec,
    QuaConfigPortReference,
    QuaConfigMultipleInputs,
    QuaConfigOctaveLoopback,
    QuaConfigCorrectionEntry,
    QuaConfigMicrowaveFemDec,
    QuaConfigMultipleOutputs,
    QuaConfigAdcPortReference,
    QuaConfigDacPortReference,
    QuaConfigPulseDecOperation,
    QuaConfigAnalogInputPortDec,
    QuaConfigDigitalWaveformDec,
    QuaConfigAnalogOutputPortDec,
    QuaConfigConstantWaveformDec,
    QuaConfigDigitalInputPortDec,
    QuaConfigOctaveLoopbackInput,
    QuaConfigOctaveLoSourceInput,
    QuaConfigOctaveRfInputConfig,
    QuaConfigArbitraryWaveformDec,
    QuaConfigDigitalOutputPortDec,
    QuaConfigGeneralPortReference,
    QuaConfigIntegrationWeightDec,
    QuaConfigOctaveRfOutputConfig,
    QuaConfigUpConverterConfigDec,
    QuaConfigDigitalWaveformSample,
    QuaConfigOctaveIfOutputsConfig,
    QuaConfigOctaveSynthesizerPort,
    QuaConfigOutputPulseParameters,
    QuaConfigSingleInputCollection,
    QuaConfigAnalogOutputPortFilter,
    QuaConfigDownConverterConfigDec,
    QuaConfigIntegrationWeightSample,
    QuaConfigOctaveOutputSwitchState,
    QuaConfigDigitalInputPortReference,
    QuaConfigDigitalOutputPortReference,
    QuaConfigOctaveSingleIfOutputConfig,
    QuaConfigOctoDacAnalogOutputPortDec,
    QuaConfigDigitalInputPortDecPolarity,
    QuaConfigMicrowaveAnalogInputPortDec,
    QuaConfigMicrowaveInputPortReference,
    QuaConfigOctaveDownconverterRfSource,
    QuaConfigOctaveSynthesizerOutputName,
    QuaConfigMicrowaveAnalogOutputPortDec,
    QuaConfigMicrowaveOutputPortReference,
    QuaConfigOutputPulseParametersPolarity,
    QuaConfigOctoDacAnalogOutputPortDecOutputMode,
    QuaConfigOctoDacAnalogOutputPortDecSamplingRate,
    QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode,
)

ALLOWED_GAINES = {x / 2 for x in range(-40, 41)}
DEFAULT_DUC_IDX = 1


def _analog_input_port_to_pb(data: AnalogInputPortConfigType) -> QuaConfigAnalogInputPortDec:
    analog_input = QuaConfigAnalogInputPortDec(
        offset=data.get("offset", 0.0),
        shareable=bool(data.get("shareable")),
        gain_db=int(data.get("gain_db", 0)),
        sampling_rate=int(data.get("sampling_rate", 1e9)),
    )
    return analog_input


def mw_fem_analog_input_port_to_pb(data: MwFemAnalogInputPortConfigType) -> QuaConfigMicrowaveAnalogInputPortDec:
    analog_input = QuaConfigMicrowaveAnalogInputPortDec(
        sampling_rate=data.get("sampling_rate", 1e9),
        gain_db=data.get("gain_db", 0),
        shareable=data.get("shareable", False),
        band=data["band"],
        downconverter=QuaConfigDownConverterConfigDec(frequency=data["downconverter_frequency"]),
    )
    return analog_input


def _get_port_reference_with_fem(reference: PortReferenceType) -> StandardPort:
    if len(reference) == 2:
        return reference[0], OPX_FEM_IDX, reference[1]
    else:
        return reference


AnalogOutputType = TypeVar("AnalogOutputType", QuaConfigAnalogOutputPortDec, QuaConfigOctoDacAnalogOutputPortDec)


def _analog_output_port_to_pb(
    data: AnalogOutputPortConfigType,
    output_type: Type[AnalogOutputType],
) -> AnalogOutputType:
    analog_output = output_type(shareable=bool(data.get("shareable")))

    if "offset" in data:
        analog_output.offset = data["offset"]

    if "delay" in data:
        delay = data.get("delay", 0)
        if delay < 0:
            raise ConfigValidationException(f"analog output delay cannot be a negative value, given value: {delay}")
        analog_output.delay = delay

    if "filter" in data:
        analog_output.filter = QuaConfigAnalogOutputPortFilter(
            feedforward=data["filter"]["feedforward"],
            feedback=data["filter"]["feedback"],
        )

    if "crosstalk" in data:
        for k, v in data["crosstalk"].items():
            analog_output.crosstalk[int(k)] = v

    return analog_output


def create_sampling_rate_enum(
    data: AnalogOutputPortConfigTypeOctoDac,
) -> QuaConfigOctoDacAnalogOutputPortDecSamplingRate:
    sampling_rate_float = data.get("sampling_rate", 1e9)
    if sampling_rate_float not in {1e9, 2e9}:
        raise ValueError("Sampling rate should be either 1e9 or 2e9")
    sampling_rate_gsps = int(sampling_rate_float / 1e9)
    return QuaConfigOctoDacAnalogOutputPortDecSamplingRate(sampling_rate_gsps)


def _opx_1000_analog_output_port_to_pb(
    data: AnalogOutputPortConfigTypeOctoDac,
) -> QuaConfigOctoDacAnalogOutputPortDec:
    item = _analog_output_port_to_pb(data, output_type=QuaConfigOctoDacAnalogOutputPortDec)
    item.sampling_rate = create_sampling_rate_enum(data)
    if item.sampling_rate == QuaConfigOctoDacAnalogOutputPortDecSamplingRate.GSPS1:
        item.upsampling_mode = QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode[data.get("upsampling_mode", "mw")]
    else:
        if "upsampling_mode" in data:
            raise ConfigValidationException("Sampling rate mode is only relevant for sampling rate of 1GHz.")
    item.output_mode = QuaConfigOctoDacAnalogOutputPortDecOutputMode[data.get("output_mode", "direct")]
    return item


def upconverter_config_dec_to_pb(
    data: Union[MwUpconverterConfigType, QuaConfigUpConverterConfigDec]
) -> QuaConfigUpConverterConfigDec:
    if isinstance(data, QuaConfigUpConverterConfigDec):
        return data
    return QuaConfigUpConverterConfigDec(frequency=data["frequency"])


def mw_fem_analog_output_to_pb(
    data: MwFemAnalogOutputPortConfigType,
) -> QuaConfigMicrowaveAnalogOutputPortDec:
    if "upconverter_frequency" in data and data.get("upconverters", {}):
        raise ConfigValidationException("Use either 'upconverter_frequency' or 'upconverters' but not both")
    if "upconverter_frequency" in data:
        upconverters = {DEFAULT_DUC_IDX: QuaConfigUpConverterConfigDec(data["upconverter_frequency"])}
    else:
        upconverters = {k: upconverter_config_dec_to_pb(v) for k, v in data.get("upconverters", {}).items()}
    if not upconverters:
        raise ConfigValidationException("You should declare at least one upconverter.")

    item = QuaConfigMicrowaveAnalogOutputPortDec(
        sampling_rate=data.get("sampling_rate", 1e9),
        full_scale_power_dbm=data.get("full_scale_power_dbm", -11),
        band=data["band"],
        delay=data.get("delay", 0),
        shareable=data.get("shareable", False),
        upconverters=upconverters,
    )
    return item


def digital_output_port_to_pb(data: DigitalOutputPortConfigType) -> QuaConfigDigitalOutputPortDec:
    digital_output = QuaConfigDigitalOutputPortDec(
        shareable=data.get("shareable", False),
        inverted=data.get("inverted", False),
        level=QuaConfigVoltageLevel[data.get("level", "LVTTL").upper()],
    )
    return digital_output


def digital_input_port_to_pb(data: DigitalInputPortConfigType) -> QuaConfigDigitalInputPortDec:
    digital_input = QuaConfigDigitalInputPortDec(shareable=bool(data.get("shareable")))

    if "threshold" in data:
        digital_input.threshold = data["threshold"]

    if "polarity" in data:
        if data["polarity"].upper() == "RISING":
            digital_input.polarity = QuaConfigDigitalInputPortDecPolarity.RISING
        elif data["polarity"].upper() == "FALLING":
            digital_input.polarity = QuaConfigDigitalInputPortDecPolarity.FALLING

    if "deadtime" in data:
        digital_input.deadtime = int(data["deadtime"])

    return digital_input


def controlling_devices_to_pb(data: Union[ControllerConfigType, OPX1000ControllerConfigType]) -> QuaConfigDeviceDec:
    fems: Dict[int, QuaConfigFemTypes] = {}

    if "fems" in data:
        data = cast(OPX1000ControllerConfigType, data)
        # Here we assume that we don't declare OPX as FEM
        if set(data) & {"analog", "analog_outputs", "digital_outputs", "digital_inputs"}:
            raise Exception(
                "'analog', 'analog_outputs', 'digital_outputs' and 'digital_inputs' are not allowed when 'fems' is present"
            )
        for k, v in data["fems"].items():
            if v.get("type") == "MW":
                fems[k] = _mw_fem_to_pb(cast(MwFemConfigType, v))
            else:
                fems[k] = _fem_to_pb(cast(LfFemConfigType, v))

    else:
        data = cast(ControllerConfigType, data)
        fems[OPX_FEM_IDX] = _controller_to_pb(data)

    item = QuaConfigDeviceDec(fems=fems)
    return item


def _controller_to_pb(data: ControllerConfigType) -> QuaConfigFemTypes:
    cont = QuaConfigControllerDec(type=data.get("type", "opx1"))
    cont = _set_ports_in_config(cont, data)
    return QuaConfigFemTypes(opx=cont)


def _fem_to_pb(data: LfFemConfigType) -> QuaConfigFemTypes:
    cont = QuaConfigOctoDacFemDec()
    cont = _set_ports_in_config(cont, data)
    return QuaConfigFemTypes(octo_dac=cont)


def _mw_fem_to_pb(data: MwFemConfigType) -> QuaConfigFemTypes:
    cont = QuaConfigMicrowaveFemDec()
    cont = _set_ports_in_config(cont, data)
    return QuaConfigFemTypes(microwave=cont)


ControllerConfigTypeVar = TypeVar(
    "ControllerConfigTypeVar", QuaConfigOctoDacFemDec, QuaConfigControllerDec, QuaConfigMicrowaveFemDec
)


def _set_ports_in_config(
    config: ControllerConfigTypeVar, data: Union[ControllerConfigType, LfFemConfigType, MwFemConfigType]
) -> ControllerConfigTypeVar:
    if "analog_outputs" in data:
        for analog_output_idx, analog_output_data in data["analog_outputs"].items():
            int_k = int(analog_output_idx)
            if isinstance(config, QuaConfigControllerDec):
                analog_output_data = cast(AnalogOutputPortConfigType, analog_output_data)
                config.analog_outputs[int_k] = _analog_output_port_to_pb(
                    analog_output_data, output_type=QuaConfigAnalogOutputPortDec
                )
            elif isinstance(config, QuaConfigOctoDacFemDec):
                analog_output_data = cast(AnalogOutputPortConfigTypeOctoDac, analog_output_data)
                config.analog_outputs[int_k] = _opx_1000_analog_output_port_to_pb(analog_output_data)
            elif isinstance(config, QuaConfigMicrowaveFemDec):
                analog_output_data = cast(MwFemAnalogOutputPortConfigType, analog_output_data)
                config.analog_outputs[int_k] = mw_fem_analog_output_to_pb(analog_output_data)
            else:
                raise ValueError(f"Unknown config type {type(config)}")

    if "analog_inputs" in data:
        if isinstance(config, (QuaConfigControllerDec, QuaConfigOctoDacFemDec)):
            for analog_input_idx, analog_input_data in data["analog_inputs"].items():
                analog_input_data = cast(AnalogInputPortConfigType, analog_input_data)
                config.analog_inputs[int(analog_input_idx)] = _analog_input_port_to_pb(analog_input_data)
                if isinstance(config, QuaConfigControllerDec):
                    sampling_rate = config.analog_inputs[int(analog_input_idx)].sampling_rate
                    if sampling_rate != 1e9:
                        raise ConfigValidationException(f"Sampling rate of {sampling_rate} is not supported for OPX")
        elif isinstance(config, QuaConfigMicrowaveFemDec):
            for analog_input_idx, analog_input_data_mw in data["analog_inputs"].items():
                analog_input_data_mw = cast(MwFemAnalogInputPortConfigType, analog_input_data_mw)
                config.analog_inputs[int(analog_input_idx)] = mw_fem_analog_input_port_to_pb(analog_input_data_mw)
        else:
            raise ValueError(f"Unknown config type {type(config)}")

    if "digital_outputs" in data:
        for digital_output_idx, digital_output_data in data["digital_outputs"].items():
            config.digital_outputs[int(digital_output_idx)] = digital_output_port_to_pb(digital_output_data)

    if "digital_inputs" in data:
        for digital_input_idx, digital_input_data in data["digital_inputs"].items():
            config.digital_inputs[int(digital_input_idx)] = digital_input_port_to_pb(digital_input_data)

    return config


def get_octave_loopbacks(data: List[LoopbackType]) -> List[QuaConfigOctaveLoopback]:
    loopbacks = [
        QuaConfigOctaveLoopback(
            lo_source_input=QuaConfigOctaveLoopbackInput[loopback[1]],
            lo_source_generator=QuaConfigOctaveSynthesizerPort(
                device_name=loopback[0][0],
                port_name=QuaConfigOctaveSynthesizerOutputName[loopback[0][1].lower()],
            ),
        )
        for loopback in data
    ]
    return loopbacks


def octave_to_pb(data: OctaveConfigType) -> QuaConfigOctaveConfig:
    connectivity = data.get("connectivity", None)
    if isinstance(connectivity, str):
        connectivity = (connectivity, OPX_FEM_IDX)
    loopbacks = get_octave_loopbacks(data.get("loopbacks", []))
    rf_modules = {
        k: rf_module_to_pb(standardize_connectivity_for_if_in(v, connectivity, k))
        for k, v in data.get("RF_outputs", {}).items()
    }
    rf_inputs = {k: rf_input_to_pb(v, k) for k, v in data.get("RF_inputs", {}).items()}
    if_outputs = _octave_if_outputs_to_pb(standardize_connectivity_for_if_out(data.get("IF_outputs", {}), connectivity))
    return QuaConfigOctaveConfig(
        loopbacks=loopbacks,
        rf_outputs=rf_modules,
        rf_inputs=rf_inputs,
        if_outputs=if_outputs,
    )


def standardize_connectivity_for_if_in(
    data: OctaveRFOutputConfigType, controller_connectivity: Optional[Tuple[str, int]], module_number: int
) -> OctaveRFOutputConfigType:
    if controller_connectivity is not None:
        if ("I_connection" in data) or ("Q_connection" in data):
            raise OctaveConnectionAmbiguity()

        data["I_connection"] = controller_connectivity + (2 * module_number - 1,)
        data["Q_connection"] = controller_connectivity + (2 * module_number,)
    return data


IF_OUT1_DEFAULT = "out1"
IF_OUT2_DEFAULT = "out2"


def standardize_connectivity_for_if_out(
    data: OctaveIfOutputsConfigType, controller_connectivity: Optional[Tuple[str, int]]
) -> OctaveIfOutputsConfigType:
    if controller_connectivity is not None:
        if "IF_out1" not in data:
            data["IF_out1"] = {"name": IF_OUT1_DEFAULT}
        if "IF_out2" not in data:
            data["IF_out2"] = {"name": IF_OUT2_DEFAULT}
        if ("port" in data["IF_out1"]) or ("port" in data["IF_out2"]):
            raise OctaveConnectionAmbiguity()
        data["IF_out1"]["port"] = controller_connectivity + (1,)
        data["IF_out2"]["port"] = controller_connectivity + (2,)
    return data


def _get_lo_frequency(data: Union[OctaveRFOutputConfigType, OctaveRFInputConfigType]) -> float:
    if "LO_frequency" not in data:
        raise ConfigValidationException("No LO frequency was set for upconverter")
    lo_freq = data["LO_frequency"]
    if not 2e9 <= lo_freq <= 18e9:
        raise ConfigValidationException(f"LO frequency {lo_freq} is out of range")
    return lo_freq


def rf_module_to_pb(data: OctaveRFOutputConfigType) -> QuaConfigOctaveRfOutputConfig:
    input_attenuators = data.get("input_attenuators", "OFF").upper()
    if input_attenuators not in {"ON", "OFF"}:
        raise ConfigValidationException("input_attenuators must be either ON or OFF")
    if "gain" not in data:
        raise ConfigValidationException("No gain was set for upconverter")
    gain = float(data["gain"])
    if gain not in ALLOWED_GAINES:
        raise ConfigValidationException(f"Gain should be an integer or half-integer between -20 and 20, got {gain})")
    to_return = QuaConfigOctaveRfOutputConfig(
        lo_frequency=_get_lo_frequency(data),
        lo_source=QuaConfigOctaveLoSourceInput[data.get("LO_source", "internal").lower()],
        output_mode=QuaConfigOctaveOutputSwitchState[data.get("output_mode", "always_off").lower()],
        gain=gain,
        input_attenuators=input_attenuators == "ON",
    )
    if "I_connection" in data:
        to_return.i_connection = dac_port_ref_to_pb(*_get_port_reference_with_fem(data["I_connection"]))
    if "Q_connection" in data:
        to_return.q_connection = dac_port_ref_to_pb(*_get_port_reference_with_fem(data["Q_connection"]))
    return to_return


def rf_input_to_pb(data: OctaveRFInputConfigType, input_idx: int = 0) -> QuaConfigOctaveRfInputConfig:
    input_idx_to_default_lo_source = {0: "not_set", 1: "internal", 2: "external"}  # 0 here is just for the default
    rf_source = QuaConfigOctaveDownconverterRfSource[data.get("RF_source", "RF_in").lower()]
    if input_idx == 1 and rf_source != QuaConfigOctaveDownconverterRfSource.rf_in:
        raise InvalidOctaveParameter("Downconverter 1 must be connected to RF-in")

    lo_source = QuaConfigOctaveLoSourceInput[data.get("LO_source", input_idx_to_default_lo_source[input_idx]).lower()]
    if input_idx == 2 and lo_source == QuaConfigOctaveLoSourceInput.internal:
        raise InvalidOctaveParameter("Downconverter 2 does not have internal LO")

    to_return = QuaConfigOctaveRfInputConfig(
        rf_source=rf_source,
        lo_frequency=_get_lo_frequency(data),
        lo_source=lo_source,
        if_mode_i=QuaConfigOctaveIfMode[data.get("IF_mode_I", "direct").lower()],
        if_mode_q=QuaConfigOctaveIfMode[data.get("IF_mode_Q", "direct").lower()],
    )
    return to_return


def single_if_output_to_pb(data: OctaveSingleIfOutputConfigType) -> QuaConfigOctaveSingleIfOutputConfig:
    controller, fem, number = _get_port_reference_with_fem(data["port"])
    return QuaConfigOctaveSingleIfOutputConfig(
        port=QuaConfigAdcPortReference(controller=controller, fem=fem, number=number), name=data["name"]
    )


def _octave_if_outputs_to_pb(data: OctaveIfOutputsConfigType) -> QuaConfigOctaveIfOutputsConfig:
    inst = QuaConfigOctaveIfOutputsConfig()
    if "IF_out1" in data:
        inst.if_out1 = single_if_output_to_pb(data["IF_out1"])
    if "IF_out2" in data:
        inst.if_out2 = single_if_output_to_pb(data["IF_out2"])
    return inst


@inject
def mixer_ref_to_pb(
    name: str,
    lo_frequency: int,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> QuaConfigMixerRef:
    item = QuaConfigMixerRef(mixer=name, lo_frequency=int(lo_frequency))
    if capabilities.supports_double_frequency:
        item.lo_frequency_double = float(lo_frequency)
    return item


@inject
def oscillator_to_pb(
    data: OscillatorConfigType, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
) -> QuaConfigOscillator:
    oscillator = QuaConfigOscillator()
    if "intermediate_frequency" in data:
        oscillator.intermediate_frequency = int(data["intermediate_frequency"])
        if capabilities.supports_double_frequency:
            oscillator.intermediate_frequency_double = float(data["intermediate_frequency"])

    if "mixer" in data:
        oscillator.mixer = QuaConfigMixerRef(mixer=data["mixer"])
        oscillator.mixer.lo_frequency = int(data.get("lo_frequency", 0))
        if capabilities.supports_double_frequency:
            oscillator.mixer.lo_frequency_double = float(data.get("lo_frequency", 0.0))

    return oscillator


@inject
def create_correction_entry(
    mixer_data: MixerConfigType,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> QuaConfigCorrectionEntry:
    correction = QuaConfigCorrectionEntry(
        frequency_negative=mixer_data["intermediate_frequency"] < 0,
        correction=QuaConfigMatrix(
            v00=mixer_data["correction"][0],
            v01=mixer_data["correction"][1],
            v10=mixer_data["correction"][2],
            v11=mixer_data["correction"][3],
        ),
    )
    correction.frequency = abs(int(mixer_data["intermediate_frequency"]))
    correction.lo_frequency = int(mixer_data["lo_frequency"])
    if capabilities.supports_double_frequency:
        correction.frequency_double = abs(float(mixer_data["intermediate_frequency"]))
        correction.lo_frequency_double = float(mixer_data["lo_frequency"])

    return correction


def mixer_to_pb(data: List[MixerConfigType]) -> QuaConfigMixerDec:
    return QuaConfigMixerDec(correction=[create_correction_entry(mixer) for mixer in data])


def element_thread_to_pb(name: str) -> QuaConfigElementThread:
    return QuaConfigElementThread(thread_name=name)


def dac_port_ref_to_pb(controller: str, fem: int, number: int) -> QuaConfigDacPortReference:
    return QuaConfigDacPortReference(controller=controller, fem=fem, number=number)


def single_input_to_pb(controller: str, fem: int, number: int) -> QuaConfigSingleInput:
    return QuaConfigSingleInput(port=dac_port_ref_to_pb(controller, fem, number))


def adc_port_ref_to_pb(controller: str, fem: int, number: int) -> QuaConfigAdcPortReference:
    return QuaConfigAdcPortReference(controller=controller, fem=fem, number=number)


def port_ref_to_pb(controller: str, fem: int, number: int) -> QuaConfigPortReference:
    return QuaConfigPortReference(controller=controller, fem=fem, number=number)


def digital_input_port_ref_to_pb(data: DigitalInputConfigType) -> QuaConfigDigitalInputPortReference:
    digital_input = QuaConfigDigitalInputPortReference(
        delay=int(data["delay"]),
        buffer=int(data["buffer"]),
    )
    if "port" in data:
        digital_input.port = port_ref_to_pb(*_get_port_reference_with_fem(data["port"]))

    return digital_input


def digital_output_port_ref_to_pb(data: PortReferenceType) -> QuaConfigDigitalOutputPortReference:
    return QuaConfigDigitalOutputPortReference(port=port_ref_to_pb(*_get_port_reference_with_fem(data)))


@inject
def element_to_pb(
    element_name: str,
    data: ElementConfigType,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> QuaConfigElementDec:
    validate_oscillator(data)
    validate_output_smearing(data)
    validate_output_tof(data)
    validate_timetagging_parameters(data)
    validate_used_inputs(data)

    element = QuaConfigElementDec()

    if "time_of_flight" in data:
        element.time_of_flight = int(data["time_of_flight"])

    if "smearing" in data:
        element.smearing = int(data["smearing"])

    if "intermediate_frequency" in data:
        element.intermediate_frequency = abs(int(data["intermediate_frequency"]))
        element.intermediate_frequency_oscillator = int(data["intermediate_frequency"])
        if capabilities.supports_double_frequency:
            element.intermediate_frequency_double = abs(float(data["intermediate_frequency"]))
            element.intermediate_frequency_oscillator_double = float(data["intermediate_frequency"])

        element.intermediate_frequency_negative = data["intermediate_frequency"] < 0

    if "thread" in data:
        element.thread = element_thread_to_pb(data["thread"])

    if "outputs" in data:
        for k, v in data["outputs"].items():
            element.outputs[k] = adc_port_ref_to_pb(*_get_port_reference_with_fem(v))
        element.multiple_outputs = QuaConfigMultipleOutputs(port_references=element.outputs)

    if "digitalInputs" in data:
        for digital_input_k, digital_input_v in data["digitalInputs"].items():
            element.digital_inputs[digital_input_k] = digital_input_port_ref_to_pb(digital_input_v)

    if "digitalOutputs" in data:
        for digital_output_k, digital_output_v in data["digitalOutputs"].items():
            element.digital_outputs[digital_output_k] = digital_output_port_ref_to_pb(digital_output_v)

    if "operations" in data:
        for op_name, op_value in data["operations"].items():
            element.operations[op_name] = op_value

    if "singleInput" in data:
        port_ref = _get_port_reference_with_fem(data["singleInput"]["port"])
        element.single_input = single_input_to_pb(*port_ref)

    if "mixInputs" in data:
        mix_inputs = data["mixInputs"]
        element.mix_inputs = QuaConfigMixInputs(
            i=dac_port_ref_to_pb(*_get_port_reference_with_fem(mix_inputs["I"])),
            q=dac_port_ref_to_pb(*_get_port_reference_with_fem(mix_inputs["Q"])),
            mixer=mix_inputs.get("mixer", ""),
        )

        lo_frequency = mix_inputs.get("lo_frequency", 0)
        element.mix_inputs.lo_frequency = int(lo_frequency)
        if capabilities.supports_double_frequency:
            element.mix_inputs.lo_frequency_double = float(lo_frequency)

    if "singleInputCollection" in data:
        element.single_input_collection = QuaConfigSingleInputCollection(
            inputs={
                k: dac_port_ref_to_pb(*_get_port_reference_with_fem(v))
                for k, v in data["singleInputCollection"]["inputs"].items()
            }
        )

    if "multipleInputs" in data:
        element.multiple_inputs = QuaConfigMultipleInputs(
            inputs={
                k: dac_port_ref_to_pb(*_get_port_reference_with_fem(v))
                for k, v in data["multipleInputs"]["inputs"].items()
            }
        )

    if "MWInput" in data:
        element.microwave_input = QuaConfigMicrowaveInputPortReference(
            port=dac_port_ref_to_pb(*_get_port_reference_with_fem(data["MWInput"]["port"])),
            upconverter=data["MWInput"].get("upconverter", DEFAULT_DUC_IDX),
        )
    if "MWOutput" in data:
        mw_output = data["MWOutput"]
        element.microwave_output = QuaConfigMicrowaveOutputPortReference(
            port=adc_port_ref_to_pb(*_get_port_reference_with_fem(mw_output["port"])),
        )

    if "oscillator" in data:
        element.named_oscillator = data["oscillator"]
    elif "intermediate_frequency" not in data:
        element.no_oscillator = Empty()

    if "sticky" in data:
        if "duration" in data["sticky"]:
            validate_sticky_duration(data["sticky"]["duration"])
        if capabilities.supports_sticky_elements:
            element.sticky = QuaConfigSticky(
                analog=data["sticky"].get("analog", True),
                digital=data["sticky"].get("digital", False),
                duration=int(data["sticky"].get("duration", 4) / 4),
            )
        else:
            if "digital" in data["sticky"] and data["sticky"]["digital"]:
                raise ConfigValidationException(
                    f"Server does not support digital sticky used in element " f"'{element_name}'"
                )
            element.hold_offset = QuaConfigHoldOffset(duration=int(data["sticky"].get("duration", 4) / 4))

    elif "hold_offset" in data:
        if capabilities.supports_sticky_elements:
            element.sticky = QuaConfigSticky(
                analog=True,
                digital=False,
                duration=data["hold_offset"].get("duration", 1),
            )
        else:
            element.hold_offset = QuaConfigHoldOffset(duration=data["hold_offset"]["duration"])

    if "outputPulseParameters" in data:
        pulse_parameters = data["outputPulseParameters"]
        output_pulse_parameters = QuaConfigOutputPulseParameters(
            signal_threshold=pulse_parameters["signalThreshold"],
        )

        signal_polarity = pulse_parameters["signalPolarity"].upper()
        if signal_polarity == "ABOVE" or signal_polarity == "ASCENDING":
            output_pulse_parameters.signal_polarity = QuaConfigOutputPulseParametersPolarity.ASCENDING
        elif signal_polarity == "BELOW" or signal_polarity == "DESCENDING":
            output_pulse_parameters.signal_polarity = QuaConfigOutputPulseParametersPolarity.DESCENDING

        if "derivativeThreshold" in pulse_parameters:
            output_pulse_parameters.derivative_threshold = pulse_parameters["derivativeThreshold"]
            polarity = pulse_parameters["derivativePolarity"].upper()
            if polarity == "ABOVE" or polarity == "ASCENDING":
                output_pulse_parameters.derivative_polarity = QuaConfigOutputPulseParametersPolarity.ASCENDING
            elif polarity == "BELOW" or polarity == "DESCENDING":
                output_pulse_parameters.derivative_polarity = QuaConfigOutputPulseParametersPolarity.DESCENDING

        element.output_pulse_parameters = output_pulse_parameters

    rf_inputs = data.get("RF_inputs", {})
    for k, (device, port) in rf_inputs.items():
        element.rf_inputs[k] = QuaConfigGeneralPortReference(device_name=device, port=port)

    rf_outputs = data.get("RF_outputs", {})
    for k, (device, port) in rf_outputs.items():
        element.rf_outputs[k] = QuaConfigGeneralPortReference(device_name=device, port=port)
    return element


def constant_waveform_to_protobuf(data: ConstantWaveFormConfigType) -> QuaConfigWaveformDec:
    return QuaConfigWaveformDec(constant=QuaConfigConstantWaveformDec(sample=data["sample"]))


def arbitrary_waveform_to_protobuf(data: ArbitraryWaveFormConfigType) -> QuaConfigWaveformDec:
    wf = QuaConfigWaveformDec()

    is_overridable = data.get("is_overridable", False)
    has_max_allowed_error = "max_allowed_error" in data
    has_sampling_rate = "sampling_rate" in data
    validate_arbitrary_waveform(is_overridable, has_max_allowed_error, has_sampling_rate)

    wf.arbitrary = QuaConfigArbitraryWaveformDec(samples=data["samples"], is_overridable=is_overridable)

    if has_max_allowed_error:
        wf.arbitrary.max_allowed_error = data["max_allowed_error"]
    elif has_sampling_rate:
        wf.arbitrary.sampling_rate = data["sampling_rate"]
    elif not is_overridable:
        wf.arbitrary.max_allowed_error = 1e-4
    return wf


def digital_waveform_to_pb(data: DigitalWaveformConfigType) -> QuaConfigDigitalWaveformDec:
    return QuaConfigDigitalWaveformDec(
        samples=[QuaConfigDigitalWaveformSample(value=bool(s[0]), length=s[1]) for s in data["samples"]]
    )


def pulse_to_pb(data: PulseConfigType) -> QuaConfigPulseDec:
    pulse = QuaConfigPulseDec()

    if "length" in data:
        pulse.length = int(data["length"])

    if data["operation"] == "control":
        pulse.operation = QuaConfigPulseDecOperation.CONTROL
    elif data["operation"] == "measurement":
        pulse.operation = QuaConfigPulseDecOperation.MEASUREMENT
    else:
        raise ConfigValidationException(f"Invalid operation {data['operation']}")

    if "digital_marker" in data:
        pulse.digital_marker = data["digital_marker"]

    if "integration_weights" in data:
        for k, v in data["integration_weights"].items():
            pulse.integration_weights[k] = v

    if "waveforms" in data:
        if "single" in data["waveforms"]:
            pulse.waveforms["single"] = data["waveforms"]["single"]

        elif "I" in data["waveforms"]:
            pulse.waveforms["I"] = data["waveforms"]["I"]
            pulse.waveforms["Q"] = data["waveforms"]["Q"]
    return pulse


def _standardize_iw_data(data: Union[List[Tuple[float, int]], List[float]]) -> List[Tuple[float, int]]:
    if len(data) == 0 or isinstance(data[0], (tuple, list)):
        to_return = []
        for x in data:
            x = cast(Tuple[float, int], x)
            to_return.append((x[0], x[1]))
        return to_return

    if isinstance(data[0], numbers.Number):
        if len(data) == 2:
            d0, d1 = cast(Tuple[float, int], data)
            return [(float(d0), int(d1))]

        data = cast(List[float], data)
        chunks = split_list_to_chunks([round(2**-15 * round(s * 2**15), 20) for s in data])
        new_data: List[Tuple[float, int]] = []
        for chunk in chunks:
            if chunk.accepts_different:
                new_data.extend([(float(u), 4) for u in chunk.data])
            else:
                new_data.append((chunk.first, 4 * len(chunk)))
        return new_data

    raise ConfigValidationException(f"Invalid IW data, data must be a list of numbers or 2-tuples, got {data}.")


def build_iw_sample(data: Union[List[Tuple[float, int]], List[float]]) -> List[QuaConfigIntegrationWeightSample]:
    clean_data = _standardize_iw_data(data)
    return [QuaConfigIntegrationWeightSample(value=s[0], length=int(s[1])) for s in clean_data]


def integration_weights_to_pb(data: IntegrationWeightConfigType) -> QuaConfigIntegrationWeightDec:
    iw = QuaConfigIntegrationWeightDec(cosine=build_iw_sample(data["cosine"]), sine=build_iw_sample(data["sine"]))
    return iw


def _all_controllers_are_opx(control_devices: Dict[str, QuaConfigDeviceDec]) -> bool:
    for device_config in control_devices.values():
        for fem_config in device_config.fems.values():
            _, controller_inst = betterproto.which_one_of(fem_config, "fem_type_one_of")
            if not isinstance(controller_inst, QuaConfigControllerDec):
                return False
    return True


def set_octave_upconverter_connection_to_elements(pb_config: QuaConfig) -> None:
    for element in pb_config.v1_beta.elements.values():
        for rf_input in element.rf_inputs.values():
            if rf_input.device_name in pb_config.v1_beta.octaves:
                if rf_input.port in pb_config.v1_beta.octaves[rf_input.device_name].rf_outputs:
                    _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
                    if element_input is not None:
                        raise ElementInputConnectionAmbiguity("Ambiguous definition of element input")

                    upconverter_config = pb_config.v1_beta.octaves[rf_input.device_name].rf_outputs[rf_input.port]
                    element.mix_inputs = QuaConfigMixInputs(
                        i=upconverter_config.i_connection, q=upconverter_config.q_connection
                    )


def _get_rf_output_for_octave(
    element: QuaConfigElementDec, octaves: Dict[str, QuaConfigOctaveConfig]
) -> Optional[QuaConfigOctaveRfOutputConfig]:
    if element.rf_inputs:
        element_rf_input = list(element.rf_inputs.values())[0]
        octave_config = octaves[element_rf_input.device_name]
        return octave_config.rf_outputs[element_rf_input.port]

    # This part is for users that do not use the rf_inputs  for connecting the octave
    element_input = element.mix_inputs
    for octave in octaves.values():
        for rf_output in octave.rf_outputs.values():
            if all(
                [
                    (rf_output.i_connection.controller == element_input.i.controller),
                    (rf_output.i_connection.fem == element_input.i.fem),
                    (rf_output.i_connection.number == element_input.i.number),
                    (rf_output.q_connection.controller == element_input.q.controller),
                    (rf_output.q_connection.fem == element_input.q.fem),
                    (rf_output.q_connection.number == element_input.q.number),
                ]
            ):
                return rf_output
    return None


@inject
def set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(
    pb_config: QuaConfig, capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
) -> None:
    for element in pb_config.v1_beta.elements.values():
        _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
        if isinstance(element_input, QuaConfigMixInputs):
            rf_output = _get_rf_output_for_octave(element, pb_config.v1_beta.octaves)
            if rf_output is None:
                continue

            if element_input.lo_frequency not in {0, int(rf_output.lo_frequency)}:
                raise ConfigValidationException(
                    "LO frequency mismatch. The frequency stated in the element is different from "
                    "the one stated in the Octave, remove the one in the element."
                )
            element_input.lo_frequency = int(rf_output.lo_frequency)
            if capabilities.supports_double_frequency:
                element_input.lo_frequency_double = rf_output.lo_frequency


I_IN_PORT = "I"
Q_IN_PORT = "Q"


def set_octave_downconverter_connection_to_elements(pb_config: QuaConfig) -> None:
    for element in pb_config.v1_beta.elements.values():
        for _, rf_output in element.rf_outputs.items():
            if rf_output.device_name in pb_config.v1_beta.octaves:
                if rf_output.port in pb_config.v1_beta.octaves[rf_output.device_name].rf_inputs:
                    downconverter_config = pb_config.v1_beta.octaves[rf_output.device_name].if_outputs
                    outputs_form_octave = {
                        downconverter_config.if_out1.name: downconverter_config.if_out1.port,
                        downconverter_config.if_out2.name: downconverter_config.if_out2.port,
                    }
                    for k, v in outputs_form_octave.items():
                        if k in element.outputs:
                            if v != element.outputs[k]:
                                raise ElementOutputConnectionAmbiguity(
                                    f"Output {k} is connected to {element.outputs[k]} but the octave "
                                    f"downconverter is connected to {v}"
                                )
                        else:
                            element.outputs[k] = v
                            _, element_outputs = betterproto.which_one_of(element, "element_outputs_one_of")
                            if isinstance(element_outputs, QuaConfigMicrowaveOutputPortReference):
                                raise ConfigValidationException("Cannot connect octave to microwave output")
                            elif isinstance(element_outputs, QuaConfigMultipleOutputs):
                                element_outputs.port_references[k] = v
                            else:
                                element.multiple_outputs = QuaConfigMultipleOutputs(port_references={k: v})


def set_non_existing_mixers_in_mix_input_elements(pb_config: QuaConfig) -> None:
    for element_name, element in pb_config.v1_beta.elements.items():
        _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
        if isinstance(element_input, QuaConfigMixInputs):
            if (
                element.intermediate_frequency
            ):  # This is here because in validation I saw that we can set an element without IF
                if not element_input.mixer:
                    element_input.mixer = f"{element_name}_mixer_{uuid.uuid4().hex[:3]}"
                    # The uuid is just to make sure the mixer doesn't exist
                if element_input.mixer not in pb_config.v1_beta.mixers:
                    pb_config.v1_beta.mixers[element_input.mixer] = QuaConfigMixerDec(
                        correction=[
                            QuaConfigCorrectionEntry(
                                frequency=element.intermediate_frequency,
                                frequency_negative=element.intermediate_frequency_negative,
                                frequency_double=element.intermediate_frequency_double,
                                lo_frequency=element_input.lo_frequency,
                                lo_frequency_double=element_input.lo_frequency_double,
                                correction=QuaConfigMatrix(v00=1, v01=0, v10=0, v11=1),
                            )
                        ]
                    )


def validate_inputs_or_outputs_exist(pb_config: QuaConfig) -> None:
    for element in pb_config.v1_beta.elements.values():
        _, element_input = betterproto.which_one_of(element, "element_inputs_one_of")
        _, element_outputs = betterproto.which_one_of(element, "element_outputs_one_of")
        if (
            element_input is None
            and element_outputs is None
            and not bool(element.outputs)  # this is for backward compatibility
            and not bool(element.digital_outputs)
            and not bool(element.digital_inputs)
        ):
            raise NoInputsOrOutputsError


def load_config_pb(config: DictQuaConfig) -> QuaConfig:
    pb_config = QuaConfig(v1_beta=QuaConfigQuaConfigV1())

    def set_controllers() -> None:
        for k, v in config["controllers"].items():
            pb_config.v1_beta.control_devices[k] = controlling_devices_to_pb(v)
        if _all_controllers_are_opx(pb_config.v1_beta.control_devices):
            for _k, _v in pb_config.v1_beta.control_devices.items():
                controller_inst = get_fem_config_instance(_v.fems[OPX_FEM_IDX])
                if not isinstance(controller_inst, QuaConfigControllerDec):
                    raise ValueError("This should not happen")
                pb_config.v1_beta.controllers[_k] = controller_inst

    def set_octaves() -> None:
        for k, v in config.get("octaves", {}).items():
            pb_config.v1_beta.octaves[k] = octave_to_pb(v)

    def set_elements() -> None:
        for k, v in config["elements"].items():
            pb_config.v1_beta.elements[k] = element_to_pb(k, v)

    def set_pulses() -> None:
        for k, v in config["pulses"].items():
            pb_config.v1_beta.pulses[k] = pulse_to_pb(v)

    def set_waveforms() -> None:
        for k, v in config["waveforms"].items():
            if v["type"] == "constant":
                pb_config.v1_beta.waveforms[k] = constant_waveform_to_protobuf(cast(ConstantWaveFormConfigType, v))
            elif v["type"] == "arbitrary":
                pb_config.v1_beta.waveforms[k] = arbitrary_waveform_to_protobuf(cast(ArbitraryWaveFormConfigType, v))
            else:
                raise ValueError("Unknown waveform type")

    def set_digital_waveforms() -> None:
        for k, v in config["digital_waveforms"].items():
            pb_config.v1_beta.digital_waveforms[k] = digital_waveform_to_pb(v)

    def set_integration_weights() -> None:
        for k, v in config["integration_weights"].items():
            pb_config.v1_beta.integration_weights[k] = integration_weights_to_pb(v)

    def set_mixers() -> None:
        for k, v in config["mixers"].items():
            pb_config.v1_beta.mixers[k] = mixer_to_pb(v)

    def set_oscillators() -> None:
        for k, v in config["oscillators"].items():
            pb_config.v1_beta.oscillators[k] = oscillator_to_pb(v)

    key_to_action = {
        "version": lambda: None,
        "controllers": set_controllers,
        "elements": set_elements,
        "pulses": set_pulses,
        "waveforms": set_waveforms,
        "digital_waveforms": set_digital_waveforms,
        "integration_weights": set_integration_weights,
        "mixers": set_mixers,
        "oscillators": set_oscillators,
        "octaves": set_octaves,
    }

    for key in config:
        key_to_action[key]()

    set_octave_upconverter_connection_to_elements(pb_config)
    set_lo_frequency_to_mix_input_elements_that_are_connected_to_octave(pb_config)
    set_octave_downconverter_connection_to_elements(pb_config)
    set_non_existing_mixers_in_mix_input_elements(pb_config)
    return pb_config



import abc
from dataclasses import dataclass
from typing import Any, List, Type, Tuple, Union, Generic, TypeVar, Optional

from dependency_injector.wiring import Provide, inject

from qm.api.models.capabilities import ServerCapabilities
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.grpc.frontend import (
    SimulationRequest,
    ExecutionRequestSimulateSimulationInterfaceNone,
    ExecutionRequestSimulateSimulationInterfaceLoopbackConnections,
    ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections,
)

SupportedConnectionTypes = Union[
    Tuple[str, int, int, str, int, int],
    Tuple[str, int, str, int],
    Tuple[str, str, int],
    Tuple[str, int, int, List[float]],
    Tuple[str, int, List[float]],
]


T = TypeVar(
    "T",
    ExecutionRequestSimulateSimulationInterfaceLoopbackConnections,
    ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections,
)


class SimulatorInterface(Generic[T], metaclass=abc.ABCMeta):
    def __init__(self, connections: List[SupportedConnectionTypes], noisePower: float = 0.0):
        self.noisePower = self._validate_and_standardize_noise_power(noisePower)
        self._connections: List[T] = self._validate_and_standardize_connections(connections)

    @abc.abstractmethod
    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        pass

    @staticmethod
    def _validate_connection_type(connection: Tuple[Any, ...], types: List[Type[Any]], expected_input: str) -> None:
        if not all(isinstance(c, t) for c, t in zip(connection, types)):
            raise Exception(f"connection should be {expected_input}")

    @staticmethod
    def _validate_and_standardize_noise_power(noise_power: float) -> float:
        if (not isinstance(noise_power, (int, float))) or noise_power < 0:
            raise Exception("noisePower must be a positive number")
        return float(noise_power)

    @classmethod
    def _validate_and_standardize_connections(cls, connections: List[SupportedConnectionTypes]) -> List[T]:
        if not isinstance(connections, list):
            raise Exception("connections argument must be of type list")
        standardized_connections = []
        for connection in connections:
            standardized_connections.append(cls._validate_and_standardize_single_connection(connection))
        return standardized_connections

    @classmethod
    @abc.abstractmethod
    def _validate_and_standardize_single_connection(cls, connection: SupportedConnectionTypes) -> T:
        pass


@inject
def get_opx_fem_number(capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]) -> int:
    """This function is here to overcome a bug in the GW, that expects the FEM to be 0"""
    return capabilities.fem_number_in_simulator


SimulationInterfaceTypes = Union[
    SimulatorInterface[ExecutionRequestSimulateSimulationInterfaceLoopbackConnections],
    SimulatorInterface[ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections],
]


class SimulationConfig:
    """Creates a configuration object to pass to
    [qm.quantum_machines_manager.QuantumMachinesManager.simulate][]

    Args:
        duration (int): The duration to run the simulation for, in clock
            cycles
        include_analog_waveforms (bool): True to collect simulated
            analog waveform names
        include_digital_waveforms (bool): True to collect simulated
            digital waveform names
        simulation_interface (SimulatorInterface):
            Interface for to simulator. Currently supported interfaces
            - ``None`` - Zero inputs
            - [qm.simulate.loopback.LoopbackInterface][] - Loopback output to input
            - [qm.simulate.raw.RawInterface][] - Specify samples for inputs
        controller_connections (List[ControllerConnection]): A list of
            connections between the controllers in the config
        extraProcessingTimeoutInMs (int): timeout in ms to wait for
            stream processing to finish. Default is -1, which is
            disables the timeout

    """

    duration = 0
    simulate_analog_outputs = False

    def __init__(
        self,
        duration: int = 0,
        include_analog_waveforms: bool = False,
        include_digital_waveforms: bool = False,
        simulation_interface: Optional[SimulationInterfaceTypes] = None,
        controller_connections: Optional[List["ControllerConnection"]] = None,
        extraProcessingTimeoutInMs: int = -1,
    ):
        if controller_connections is None:
            controller_connections = []
        super(SimulationConfig, self).__init__()
        self.duration = duration
        self.include_analog_waveforms = include_analog_waveforms is True
        self.include_digital_waveforms = include_digital_waveforms is True
        self.simulation_interface = simulation_interface
        self.controller_connections = controller_connections
        self.extraProcessingTimeoutInMs = extraProcessingTimeoutInMs

    def update_simulate_request(self, request: SimulationRequest) -> SimulationRequest:
        if self.simulation_interface is None:
            request.simulate.simulation_interface.none = ExecutionRequestSimulateSimulationInterfaceNone()
        else:
            request = self.simulation_interface.update_simulate_request(request)
        return request


@dataclass
class InterOpxAddress:
    """Args:
    controller (str): The name of the controller
    is_left_connection (bool)
    """

    controller: str
    is_left_connection: bool


@dataclass
class InterOpxChannel:
    """Args:
    controller (str): The name of the controller
    channel_number (int)
    """

    controller: str
    channel_number: int


InterOpxPairing = Union[InterOpxAddress, InterOpxChannel]


@dataclass
class ControllerConnection:
    source: InterOpxPairing
    target: InterOpxPairing



from qm.exceptions import QmQuaException
from qm.qua._dsl import _Variable as Variable  # noqa
from qm.qua.AnalogMeasureProcess import AnalogMeasureProcess
from qm.qua.DigitalMeasureProcess import DigitalMeasureProcess
from qm.qua.lib import Cast, Math, Util, Random, call_library_function  # noqa
from qm.qua._expressions import IO1, IO2, AllQuaTypes, QuaNumberType, QuaVariableType, QuaExpressionType  # noqa
from qm.qua._dsl_specific_type_hints import ChirpType, AmpValuesType, PlayPulseType, MeasurePulseType, TypeOrExpression
from qm.qua._type_hinting import (
    OneOrMore,
    AllPyTypes,
    MessageVarType,
    ForEachValuesType,
    PyNumberArrayType,
    MessageExpressionType,
    VariableDeclarationType,
    MessageVariableOrExpression,
    fixed,
)
from qm.qua._dsl import (
    for_each_,
    dual_demod,
    integration,
    reset_frame,
    reset_phase,
    ramp_to_zero,
    time_tagging,
    set_dc_offset,
    declare_stream,
    frame_rotation,
    infinite_loop_,
    port_condition,
    reset_if_phase,
    strict_timing_,
    dual_integration,
    update_frequency,
    wait_for_trigger,
    stream_processing,
    update_correction,
    frame_rotation_2pi,
    reset_global_phase,
    fast_frame_rotation,
    advance_input_stream,
    declare_input_stream,
)
from qm.qua._dsl import (
    FUNCTIONS,
    L,
    Counting,
    StreamType,
    TimeTagging,
    CommandsType,
    DeclarationType,
    TimeDivisionType,
    AccumulationMethod,
    MeasureProcessType,
    DualAccumulationMethod,
    RealAccumulationMethod,
    AnalogProcessTargetType,
    amp,
    exp,
    if_,
    bins,
    for_,
    play,
    ramp,
    save,
    wait,
    align,
    case_,
    demod,
    elif_,
    else_,
    pause,
    assign,
    while_,
    declare,
    measure,
    program,
    switch_,
    counting,
    default_,
)

__all__ = [
    "L",
    "if_",
    "amp",
    "exp",
    "IO1",
    "IO2",
    "Math",
    "Cast",
    "Util",
    "play",
    "wait",
    "save",
    "for_",
    "ramp",
    "bins",
    "pause",
    "align",
    "case_",
    "else_",
    "elif_",
    "fixed",
    "demod",
    "Random",
    "assign",
    "while_",
    "program",
    "measure",
    "switch_",
    "declare",
    "default_",
    "Counting",
    "counting",
    "Variable",
    "ChirpType",
    "OneOrMore",
    "for_each_",
    "FUNCTIONS",
    "AllPyTypes",
    "StreamType",
    "dual_demod",
    "reset_frame",
    "reset_phase",
    "TimeTagging",
    "integration",
    "AllQuaTypes",
    "ramp_to_zero",
    "time_tagging",
    "CommandsType",
    "AmpValuesType",
    "PlayPulseType",
    "QuaNumberType",
    "set_dc_offset",
    "QuaNumberType",
    "reset_if_phase",
    "MessageVarType",
    "frame_rotation",
    "QmQuaException",
    "infinite_loop_",
    "port_condition",
    "declare_stream",
    "strict_timing_",
    "DeclarationType",
    "QuaVariableType",
    "MeasurePulseType",
    "TypeOrExpression",
    "TimeDivisionType",
    "update_frequency",
    "wait_for_trigger",
    "dual_integration",
    "ForEachValuesType",
    "PyNumberArrayType",
    "update_correction",
    "stream_processing",
    "QuaExpressionType",
    "MeasureProcessType",
    "frame_rotation_2pi",
    "reset_global_phase",
    "AccumulationMethod",
    "fast_frame_rotation",
    "AnalogMeasureProcess",
    "declare_input_stream",
    "advance_input_stream",
    "DigitalMeasureProcess",
    "MessageExpressionType",
    "call_library_function",
    "RealAccumulationMethod",
    "DualAccumulationMethod",
    "AnalogProcessTargetType",
    "VariableDeclarationType",
    "MessageVariableOrExpression",
]



from qm.simulate.loopback import LoopbackInterface  # noqa
from qm.simulate.interface import (  # noqa
    InterOpxAddress,
    InterOpxChannel,
    InterOpxPairing,
    SimulationConfig,
    ControllerConnection,
)

__all__ = [
    "SimulationConfig",
    "InterOpxAddress",
    "InterOpxChannel",
    "ControllerConnection",
    "InterOpxPairing",
    "LoopbackInterface",
]



import abc
import time
import logging
from enum import Enum, auto
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, Mapping, Optional

import numpy as np
from octave_sdk import Octave, RFInputLOSource
from octave_sdk._errors import InvalidLoSource
from octave_sdk._octave_client import OctaveClient
from octave_sdk.octave import UnableToSetFrequencyError
from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    SynthUpdate,
    OctaveModule,
    SingleUpdate,
    RfUpConvUpdate,
    ModuleReference,
    IfDownConvUpdate,
    RfDownConvUpdate,
    IfDownConvUpdateMode,
    SynthUpdateMainOutput,
    IfDownConvUpdateChannel,
    RfDownConvUpdateLoInput,
    RfDownConvUpdateRfInput,
    IfDownConvUpdateCoupling,
    SynthUpdateSecondaryOutput,
    SynthUpdateSynthOutputPower,
    RfUpConvUpdateFastSwitchMode,
)

from qm.program import Program
from qm.api.v2.job_api import JobApi
from qm.elements.element import Element
from qm.jobs.running_qm_job import RunningQmJob
from qm.grpc.qua_config import QuaConfigMixInputs
from qm.elements.up_converted_input import UpconvertedInput
from qm.api.v2.job_api.element_input_api import MixInputsApi
from qm.exceptions import QmQuaException, CantCalibrateElementError
from qm.octave._calibration_names import SavedVariablesNames, CalibrationElementsNames
from qm.octave._calibration_analysis import (
    Array,
    LOAnalysisDebugData,
    ImageDataAnalysisResult,
    _get_and_analyze_lo_data,
    _get_and_analyze_image_data,
)
from qm.qua import (
    IO1,
    IO2,
    Util,
    amp,
    if_,
    for_,
    play,
    save,
    align,
    else_,
    pause,
    assign,
    while_,
    declare,
    measure,
    program,
    dual_demod,
    reset_if_phase,
    update_correction,
)

logger = logging.getLogger(__name__)

GainType = Optional[float]
FreqType = Union[int, float]

if TYPE_CHECKING:
    from qm.api.v2.qm_api import QmApi
    from qm.quantum_machine import QuantumMachine

CALIBRATION_INPUT = 2
OTHER_INPUT = 1


class DConvQuadrature(Enum):
    I = auto()
    Q = auto()
    IQ = auto()


@dataclass
class LOFrequencyDebugData:
    """
    Debug data for the LO frequency calibration.
    prev_result: The previous result of the calibration.
    power_amp_attn: The power amplifier attenuation.
    coarse: The debug data for the coarse scan.
    fine: The debug data for the fine scan.
    """

    prev_result: Optional[Tuple[float, ...]]
    power_amp_attn: Optional[int]
    coarse: List[LOAnalysisDebugData]
    fine: List[LOAnalysisDebugData]


@dataclass
class ImageResult:
    """
    Result of the image data calibration.
    prev_result: The previous result of the calibration.
    coarse: The coarse scan result.
    fine: The fine scan result.
    """

    prev_result: Optional[Tuple[float, float]]
    coarse: ImageDataAnalysisResult
    fine: ImageDataAnalysisResult


@dataclass
class LOFrequencyCalibrationResult:
    """
    Calibration result for a single LO frequency and its corresponding IF frequencies.
    i0: The I offset.
    q0: The Q offset.
    dc_gain: The gain used during the calibration.
    dc_phase: The phase used during the calibration.
    temperature: The temperature during the calibration.
    image: The calibration results for the IF frequencies.
    debug: The debug data.
    plugin_data: Any additional data.
    """

    i0: float
    q0: float
    dc_gain: float
    dc_phase: float
    temperature: float
    image: Dict[FreqType, ImageResult]
    debug: LOFrequencyDebugData
    plugin_data: Any = None


MixerCalibrationResults = Dict[Tuple[FreqType, GainType], LOFrequencyCalibrationResult]


@dataclass
class DeprecatedCalibrationResult:
    correction: List[float]
    i_offset: float
    q_offset: float
    lo_frequency: float
    if_frequency: float
    temperature: float
    mixer_id: str
    optimizer_parameters: Dict[str, Any]


DeprecatedMixerCalibrationResults = Dict[Tuple[int, int], DeprecatedCalibrationResult]


class CalibrationCallback(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __call__(self, lo_freq: Union[int, float], lo_calibration_result: LOFrequencyCalibrationResult) -> Any:
        pass


@dataclass
class AutoCalibrationParams:
    use_main_lo: bool = False
    external_loopback: bool = False
    offset_frequency: float = 7e6
    dconv_quadrature: DConvQuadrature = DConvQuadrature.IQ
    if_amplitude: float = 0.125
    calibrate_lo_with_if: bool = True
    callback: Optional[CalibrationCallback] = None


class NoCalibrationElements(QmQuaException):
    pass


@dataclass
class _StateRestoreParams:
    octave_updates: List[SingleUpdate]
    lo_frequency: float
    i_offset: float
    q_offset: float
    input1_offset: float
    input2_offset: float


def convert_to_old_calibration_result(
    calibration_result: MixerCalibrationResults, mixer_id: str
) -> DeprecatedMixerCalibrationResults:
    to_return = {}
    for (lo_freq, _), lo_result in calibration_result.items():
        for if_freq, if_result in lo_result.image.items():
            curr_result = DeprecatedCalibrationResult(
                correction=list(if_result.fine.correction),
                i_offset=lo_result.i0,
                q_offset=lo_result.q0,
                lo_frequency=lo_freq,
                if_frequency=if_freq,
                temperature=lo_result.temperature,
                mixer_id=mixer_id,
                optimizer_parameters={},
            )
            to_return[(int(lo_freq), int(if_freq))] = curr_result
    return to_return


class OctaveMixerCalibrationBase(metaclass=abc.ABCMeta):
    def __init__(self, client: Octave):
        self._client = client

        self._lo_step = 2**-3
        self._lo_range = 2**1
        self._n_lo = int(self._lo_range / self._lo_step) + 1

        self._image_step = 2**-6
        self._image_range = 2**-1
        self._n_image = int(self._image_range / self._image_step) + 1

        self._names: Optional[CalibrationElementsNames] = None

    @property
    def names(self) -> CalibrationElementsNames:
        if self._names is None:
            raise ValueError("The calibration elements names were not set.")
        return self._names

    @property
    def _low_level_client(self) -> OctaveClient:
        return self._client._client

    def _generate_program(self, params: AutoCalibrationParams) -> Program:
        elements_names = self.names

        def read_lo_and_image_power() -> None:
            align()

            reset_if_phase(elements_names.iq_mixer)
            reset_if_phase(elements_names.lo_analyzer)
            reset_if_phase(elements_names.image_analyzer)

            play("DC_offset" * amp(i0_curr), elements_names.i_offset)
            play("DC_offset" * amp(q0_curr), elements_names.q_offset)
            play("calibration" * amp(scale_cal), elements_names.iq_mixer)

            if params.dconv_quadrature == DConvQuadrature.I:
                measure(
                    "Analyze",
                    elements_names.lo_analyzer,
                    None,
                    dual_demod.full("integW_cos", "out1", "integW_zero", "out2", I2),
                    dual_demod.full("integW_minus_sin", "out1", "integW_zero", "out2", Q2),
                )
                measure(
                    "Analyze",
                    elements_names.image_analyzer,
                    None,
                    dual_demod.full("integW_cos", "out1", "integW_zero", "out2", I3),
                    dual_demod.full("integW_minus_sin", "out1", "integW_zero", "out2", Q3),
                )
            elif params.dconv_quadrature == DConvQuadrature.Q:
                measure(
                    "Analyze",
                    elements_names.lo_analyzer,
                    None,
                    dual_demod.full("integW_zero", "out1", "integW_sin", "out2", I2),
                    dual_demod.full("integW_zero", "out1", "integW_cos", "out2", Q2),
                )
                measure(
                    "Analyze",
                    elements_names.image_analyzer,
                    None,
                    dual_demod.full("integW_zero", "out1", "integW_sin", "out2", I3),
                    dual_demod.full("integW_zero", "out1", "integW_cos", "out2", Q3),
                )
            else:
                measure(
                    "Analyze",
                    elements_names.lo_analyzer,
                    None,
                    dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I2),
                    dual_demod.full("integW_minus_sin", "out1", "integW_cos", "out2", Q2),
                )
                measure(
                    "Analyze",
                    elements_names.image_analyzer,
                    None,
                    dual_demod.full("integW_cos", "out1", "integW_sin", "out2", I3),
                    dual_demod.full("integW_minus_sin", "out1", "integW_cos", "out2", Q3),
                )

            assign(lo_power, (I2 * I2 + Q2 * Q2))
            assign(image_power, (I3 * I3 + Q3 * Q3))

        def update_correction_g_phi() -> None:
            assign(s_mat, p_curr)
            assign(s2, s_mat * s_mat)
            assign(c_mat, 1 + 1.5 * s2 - 3.125 * (s2 * s2))
            assign(g_mat_plus, 1 + g_curr + 0.5 * g_curr * g_curr)
            assign(g_mat_minus, 1 - g_curr + 0.5 * g_curr * g_curr)
            assign(c00, g_mat_plus * c_mat)
            assign(c01, g_mat_plus * s_mat)
            assign(c10, g_mat_minus * s_mat)
            assign(c11, g_mat_minus * c_mat)
            update_correction(elements_names.iq_mixer, c00, c01, c10, c11)

        with program() as prog:

            # variables for getting IQ data
            Q2 = declare(float)
            Q3 = declare(float)
            I2 = declare(float)
            I3 = declare(float)

            lo_power = declare(float)
            image_power = declare(float)

            i0_center = declare(float)
            q0_center = declare(float)
            i0_offset = declare(float)
            q0_offset = declare(float)
            i0_curr = declare(float)
            q0_curr = declare(float)
            i0_best = declare(float)
            q0_best = declare(float)

            scale_iq = declare(float)
            scale_gp = declare(float)
            scale_cal = declare(float)

            g_center = declare(float)
            p_center = declare(float)
            g_offset = declare(float)
            p_offset = declare(float)
            g_curr = declare(float)
            p_curr = declare(float)

            # variables for correction matrix
            g_mat_plus = declare(float)
            g_mat_minus = declare(float)
            s2 = declare(float)
            c_mat = declare(float)
            s_mat = declare(float)
            c00 = declare(float)
            c01 = declare(float)
            c10 = declare(float)
            c11 = declare(float)

            best_lo_power = declare(float)

            task = declare(int)
            go_on = declare(bool, value=True)

            with while_(go_on):

                pause()
                assign(task, IO1)

                with if_(task < 2):

                    assign(scale_cal, 1.0 if params.calibrate_lo_with_if else 0.0)
                    assign(scale_iq, Util.cond(task == 0, 1.0, 0.125))
                    assign(i0_center, Util.cond(task == 0, 0.0, i0_center))
                    assign(q0_center, Util.cond(task == 0, 0.0, q0_center))

                    # # I-Q scans are done with no mixer correction!
                    # update_correction(elements_names.iq_mixer, 1.0, 0.0, 0.0, 1.0)

                    assign(best_lo_power, 7.96875)
                    with for_(
                        i0_offset, -self._lo_range / 2, i0_offset <= self._lo_range / 2, i0_offset + self._lo_step
                    ):
                        with for_(
                            q0_offset, -self._lo_range / 2, q0_offset <= self._lo_range / 2, q0_offset + self._lo_step
                        ):
                            assign(i0_curr, i0_offset * scale_iq + i0_center)
                            assign(q0_curr, q0_offset * scale_iq + q0_center)
                            read_lo_and_image_power()

                            with if_(lo_power < best_lo_power):
                                assign(i0_best, i0_curr)
                                assign(q0_best, q0_curr)
                                assign(best_lo_power, lo_power)

                            save(i0_curr, SavedVariablesNames.i_scan)
                            save(q0_curr, SavedVariablesNames.q_scan)

                            save(lo_power, SavedVariablesNames.lo)

                    assign(i0_center, Util.cond(task == 1, i0_best, 0.0))
                    assign(q0_center, Util.cond(task == 1, q0_best, 0.0))

                    with if_(task == 0):
                        measure("Analyze", elements_names.lo_analyzer, "lo_adc_data")

                with else_():

                    with if_(task < 4):

                        assign(scale_cal, 1.0)
                        assign(scale_gp, Util.cond(task == 2, params.if_amplitude * 8, params.if_amplitude))
                        assign(i0_curr, 0.0)
                        assign(q0_curr, 0.0)

                        pause()
                        assign(g_center, IO1)
                        assign(p_center, IO2)

                        with for_(
                            g_offset,
                            -self._image_range / 2,
                            g_offset <= self._image_range / 2,
                            g_offset + self._image_step,
                        ):
                            with for_(
                                p_offset,
                                -self._image_range / 2,
                                p_offset <= self._image_range / 2,
                                p_offset + self._image_step,
                            ):
                                assign(g_curr, g_offset * scale_gp + g_center)
                                assign(p_curr, p_offset * scale_gp + p_center)

                                update_correction_g_phi()

                                read_lo_and_image_power()

                                save(g_curr, SavedVariablesNames.g_scan)
                                save(p_curr, SavedVariablesNames.p_scan)

                                save(image_power, SavedVariablesNames.image)

                    with else_():
                        assign(go_on, False)

        return prog

    def _set_octave_for_calibration(
        self, output_channel_index: int, params: AutoCalibrationParams, element_input: UpconvertedInput
    ) -> _StateRestoreParams:
        current_state = self._low_level_client.aquire_modules(
            modules=[
                ModuleReference(type=OctaveModule.RF_UPCONVERTER, index=1),
                ModuleReference(type=OctaveModule.RF_UPCONVERTER, index=2),
                ModuleReference(type=OctaveModule.RF_UPCONVERTER, index=3),
                ModuleReference(type=OctaveModule.RF_UPCONVERTER, index=4),
                ModuleReference(type=OctaveModule.RF_UPCONVERTER, index=5),
                ModuleReference(type=OctaveModule.RF_DOWNCONVERTER, index=CALIBRATION_INPUT),
                ModuleReference(type=OctaveModule.IF_DOWNCONVERTER, index=CALIBRATION_INPUT),
                ModuleReference(type=OctaveModule.SYNTHESIZER, index=4),
                ModuleReference(type=OctaveModule.RF_DOWNCONVERTER, index=OTHER_INPUT),
                ModuleReference(type=OctaveModule.IF_DOWNCONVERTER, index=OTHER_INPUT),
            ]
        ).state.updates

        state_restore_updates = [
            current_state[5],  # take RF-downconverter #2 as is
            current_state[6],  # take IF-downconverter #2 as is
            current_state[7],  # take SYNTH #4 as is (the calibration synth)
        ]
        for upconverter_index in (1, 2, 3, 4, 5):
            upconverter_state = current_state[upconverter_index - 1].rf_up_conv
            state_restore_updates.append(
                SingleUpdate(
                    rf_up_conv=RfUpConvUpdate(
                        index=upconverter_index,
                        enabled=upconverter_state.enabled,
                        mixer_output_attn=upconverter_state.mixer_output_attn,
                        power_amp_attn=upconverter_state.power_amp_attn,
                        fast_switch_mode=upconverter_state.fast_switch_mode,
                    )
                )
            )

        restore_lo_frequency = element_input.lo_frequency

        if params.external_loopback:
            rf_input = RfDownConvUpdateRfInput.RF_INPUT_MAIN
        else:
            rf_input = {
                1: RfDownConvUpdateRfInput.RF_INPUT_DEBUG_1,
                2: RfDownConvUpdateRfInput.RF_INPUT_DEBUG_2,
                3: RfDownConvUpdateRfInput.RF_INPUT_DEBUG_3,
                4: RfDownConvUpdateRfInput.RF_INPUT_DEBUG_4,
                5: RfDownConvUpdateRfInput.RF_INPUT_DEBUG_5,
            }[output_channel_index]

        # Shut down all the over up-converters while doing the calibration
        updates = [
            SingleUpdate(
                rf_down_conv=RfDownConvUpdate(
                    index=CALIBRATION_INPUT,
                    enabled=True,
                    lo_input=RfDownConvUpdateLoInput.LO_INPUT_1
                    if params.use_main_lo
                    else RfDownConvUpdateLoInput.LO_INPUT_2,
                    rf_input=rf_input,
                )
            ),
            SingleUpdate(
                rf_down_conv=RfDownConvUpdate(
                    index=OTHER_INPUT,
                    enabled=False,
                    lo_input=RfDownConvUpdateLoInput.LO_INPUT_1,
                    rf_input=RfDownConvUpdateRfInput.RF_INPUT_DISCONNECT,
                )
            ),
            SingleUpdate(
                synth=SynthUpdate(
                    index=4,
                    gain=0xFFFF,
                    synth_output_power=SynthUpdateSynthOutputPower.SYNTH_OUTPUT_POWER_POS5DB,
                    main_output=SynthUpdateMainOutput.MAIN_OUTPUT_MAIN
                    if params.use_main_lo
                    else SynthUpdateMainOutput.MAIN_OUTPUT_OFF,
                    secondary_output=SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_OFF,
                )
            ),
            SingleUpdate(
                if_down_conv=IfDownConvUpdate(
                    index=CALIBRATION_INPUT,
                    channel1=IfDownConvUpdateChannel(
                        mode=IfDownConvUpdateMode.MODE_BYPASS
                        if params.dconv_quadrature != DConvQuadrature.I
                        else IfDownConvUpdateMode.MODE_OFF,
                        coupling=IfDownConvUpdateCoupling.COUPLING_AC,
                    ),
                    channel2=IfDownConvUpdateChannel(
                        mode=IfDownConvUpdateMode.MODE_BYPASS
                        if params.dconv_quadrature != DConvQuadrature.Q
                        else IfDownConvUpdateMode.MODE_OFF,
                        coupling=IfDownConvUpdateCoupling.COUPLING_AC,
                    ),
                )
            ),
            SingleUpdate(
                if_down_conv=IfDownConvUpdate(
                    index=OTHER_INPUT,
                    channel1=IfDownConvUpdateChannel(
                        mode=IfDownConvUpdateMode.MODE_OFF,
                        coupling=IfDownConvUpdateCoupling.COUPLING_AC,
                    ),
                    channel2=IfDownConvUpdateChannel(
                        mode=IfDownConvUpdateMode.MODE_OFF,
                        coupling=IfDownConvUpdateCoupling.COUPLING_AC,
                    ),
                )
            ),
        ]
        for index in (1, 2, 3, 4, 5):
            if index != output_channel_index:
                updates.append(
                    SingleUpdate(
                        rf_up_conv=RfUpConvUpdate(
                            index=index,
                            mixer_output_attn=63,
                            power_amp_attn=63,
                            fast_switch_mode=RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_OFF,
                        )
                    )
                )
            else:
                updates.append(
                    SingleUpdate(
                        rf_up_conv=RfUpConvUpdate(
                            index=index,
                            fast_switch_mode=RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_OFF,
                            enabled=True
                            # YR - I added the enabled update to make the assertion down the code pass
                        )
                    )
                )

        self._low_level_client.update(updates=updates)
        i_offset, q_offset, input1_offset, input2_offset = self._get_element_idle_offsets(element_input._name)
        restore_params = _StateRestoreParams(
            octave_updates=state_restore_updates,
            lo_frequency=restore_lo_frequency,
            i_offset=i_offset,
            q_offset=q_offset,
            input1_offset=input1_offset,
            input2_offset=input2_offset,
        )
        return restore_params

    @staticmethod
    def _set_input_lo_frequency(element_input: UpconvertedInput, lo_freq: FreqType) -> None:
        """Even if the LO source is external, we will have to update the attenuators."""
        try:
            element_input.set_lo_frequency(lo_freq, set_source=False)
        except (UnableToSetFrequencyError, InvalidLoSource):
            element_input.inform_element_about_lo_frequency(lo_freq)
            if element_input.lo_frequency != lo_freq:
                logger.warning(
                    "Could not set the external LO frequency to the desired value, "
                    "make sure to calibrate the declared LO frequency."
                )

    def _set_before_coarse_scan(
        self,
        element_input: UpconvertedInput,
        lo_freq: FreqType,
        offset_frequency: float,
    ) -> None:

        self._set_input_lo_frequency(element_input, lo_freq)
        self._client.rf_inputs[2].set_lo_source(source_name=RFInputLOSource.Internal, ignore_shared_errors=True)
        self._client.rf_inputs[2].set_lo_frequency(
            source_name=RFInputLOSource.Internal, frequency=lo_freq + offset_frequency
        )

        self._set_iq_mixer_element_input_dc_offsets(i_offset=0.0, q_offset=0.0)
        self._set_if_freq(50.0e6, down_mixer_offset=offset_frequency)

    def _perform_coarse_iq_scan(
        self, job: Union[RunningQmJob, JobApi], n_lo: int, lo_offset: int
    ) -> Tuple[Array, Array, List[LOAnalysisDebugData], Array]:
        self._set_io_values(io1=0)

        job.resume()
        i0_coarse, q0_coarse, debug_coarse = _get_and_analyze_lo_data(job, n_lo, lo_offset, 1)

        lo_adc_data_input1 = job.result_handles.get("lo_adc_data_input1")
        lo_adc_data_input2 = job.result_handles.get("lo_adc_data_input2")
        assert lo_adc_data_input1 is not None
        assert lo_adc_data_input2 is not None

        lo_adc_data_input1.wait_for_values(1)
        lo_adc_data_input2.wait_for_values(1)
        lo_adc_data = np.array(
            [
                lo_adc_data_input1.fetch(slice(0, 1), flat_struct=True),
                lo_adc_data_input2.fetch(slice(0, 1), flat_struct=True),
            ]
        )
        return i0_coarse, q0_coarse, debug_coarse, lo_adc_data

    def _perform_fine_iq_scan(
        self,
        job: Union[RunningQmJob, JobApi],
        n_lo: int,
        lo_offset: int,
    ) -> Tuple[Array, Array, List[LOAnalysisDebugData]]:
        self._set_io_values(io1=1)

        for _ in range(2):
            job.resume()
            while not job.is_paused():
                time.sleep(0.001)

        i0_shift, q0_shift, debug_fine = _get_and_analyze_lo_data(job, n_lo, lo_offset, 2)
        return i0_shift, q0_shift, debug_fine

    def _restore_previous_state(self, restore_params: _StateRestoreParams, element_input: UpconvertedInput) -> None:
        self._low_level_client.update(updates=restore_params.octave_updates)
        self._set_input_lo_frequency(element_input, restore_params.lo_frequency)

    def calibrate(
        self,
        element: Element[QuaConfigMixInputs],
        lo_if_dict: Mapping[float, Tuple[float, ...]],
        params: AutoCalibrationParams,
    ) -> MixerCalibrationResults:
        element_input = element.input
        if not isinstance(element_input, UpconvertedInput):
            raise CantCalibrateElementError(
                f"Element {element.name} has input of type {type(element.input)} and is not connected to an Octave. "
                f"Hence, it cannot be calibrated."
            )

        t_start = time.time()
        octave = self._client
        octave_port = element_input.port

        port_idx = octave_port[1]
        self._names = CalibrationElementsNames(self._low_level_client.octave_name, port_idx)

        restore_params = self._set_octave_for_calibration(port_idx, params, element_input)

        compiled = self._compile_program(self._generate_program(params))

        n_lo_samples, n_image_samples = self._n_lo, self._n_image

        # Now we go through the LO frequencies. For each LO we calibrate the DC
        # offsets, and then calibrate the image for each if freq.
        result = {}
        start_lo_sweep = time.time()
        logger.debug(f"time to start LO sweep {start_lo_sweep - t_start}")

        for lo_freq in lo_if_dict:
            t0_lo = time.time()
            logger.debug(f"Calibrating {lo_freq / 1e9:0.3f} GHz")

            job = self._execute_job(compiled)

            lo_offset, image_offset = 0, 0

            while not job.is_paused():
                time.sleep(0.01)

            # Let's look if there is a previous calibration value (to compate with at the end)
            prev_lo_cal = None

            logger.debug(f"Calibrating {lo_freq / 1e9:0.3f} GHz")
            self._set_before_coarse_scan(element_input, lo_freq, params.offset_frequency)

            self._set_analyzer_element_outputs_dc_offsets(out1=0.0, out2=0.0)

            i0_coarse, q0_coarse, debug_coarse, lo_adc_data = self._perform_coarse_iq_scan(job, n_lo_samples, lo_offset)
            lo_offset += 1

            self._set_analyzer_element_outputs_dc_offsets(
                out1=-np.mean(lo_adc_data[0]) / 4096, out2=-np.mean(lo_adc_data[1]) / 4096
            )

            # The first scan is coarse and use to find a good starting point
            self._set_iq_mixer_element_input_dc_offsets(i_offset=i0_coarse[0], q_offset=q0_coarse[0])

            i0_shift, q0_shift, debug_fine = self._perform_fine_iq_scan(job, n_lo_samples, lo_offset)
            lo_offset += 2

            curr_lo_freq_result = LOFrequencyCalibrationResult(
                i0=i0_shift[0] + i0_coarse[0],
                q0=q0_shift[0] + q0_coarse[0],
                dc_gain=debug_coarse[0].corrections.dc_gain,
                dc_phase=debug_coarse[0].corrections.dc_phase,
                temperature=octave.rf_outputs[port_idx].get_temperature(),
                image={},
                debug=LOFrequencyDebugData(
                    prev_result=prev_lo_cal, power_amp_attn=None, coarse=debug_coarse, fine=debug_fine
                ),
            )

            self._set_iq_mixer_element_input_dc_offsets(
                i_offset=curr_lo_freq_result.i0, q_offset=curr_lo_freq_result.q0
            )

            logger.debug(f"Pre IF sweep time - {time.time() - t0_lo}")
            for if_freq in lo_if_dict[lo_freq]:
                prev_if_cal = None

                t0_if = time.time()
                logger.debug(f"if_freq = {if_freq / 1e6:0.1f} MHz")

                self._set_if_freq(float(if_freq), params.offset_frequency)

                # 2 - image scan
                self._set_io_values(io1=2)
                job.resume()
                #  Set the initial guess for the gain and phase from the coarse DC scan
                coarse_gain, coarse_phase = curr_lo_freq_result.dc_gain, curr_lo_freq_result.dc_phase
                self._set_io_values(io1=coarse_gain, io2=coarse_phase)
                logger.debug(f"coarse. gain = {coarse_gain:.5f}, phase = {coarse_phase:0.5f}")
                while not job.is_paused():
                    time.sleep(0.001)
                job.resume()

                gp_coarse = _get_and_analyze_image_data(job, n_image_samples, image_offset, 1)[0]
                image_offset += 1

                # prepare for fine scan
                self._set_io_values(io1=3)
                job.resume()

                # Set the initial guess for the gain and phase from the coarse gate-phase scan
                coarse_gain, coarse_phase = gp_coarse.gain, gp_coarse.phase
                self._set_io_values(io1=coarse_gain, io2=coarse_phase)
                logger.debug(f"coarse. gain = {coarse_gain:.5f}, phase = {coarse_phase:0.5f}")
                while not job.is_paused():
                    time.sleep(0.001)
                job.resume()

                gp_fine = _get_and_analyze_image_data(job, n_image_samples, image_offset, 1)[0]

                if abs(gp_fine.gain) > 0.3 or abs(gp_fine.phase) > 0.3:
                    logger.debug(
                        f"Failed to calibrate LO={lo_freq / 1e9:0.3f}GHz, IF={if_freq / 1e6:0.3f}MHz, "
                        f"setting to identity"
                    )
                    gp_fine.gain = 0.0
                    gp_fine.phase = 0.0

                curr_lo_freq_result.image[if_freq] = ImageResult(
                    prev_result=prev_if_cal, coarse=gp_coarse, fine=gp_fine
                )
                image_offset += 1
                logger.debug(f"fine. gain = {gp_fine.gain:.5f}, phase = {gp_fine.phase:0.5f}")

                logger.debug(f"Calibration for {lo_freq}, {if_freq} took {time.time() - t0_if}")

            result[(lo_freq, element_input.gain)] = curr_lo_freq_result

            self._set_io_values(io1=4)
            job.resume()

            if params.callback is not None:
                curr_lo_freq_result.plugin_data = params.callback(lo_freq, curr_lo_freq_result)
            logger.debug(f"Calibration for LO {lo_freq} took {time.time() - t0_lo}")

        self._restore_previous_state(restore_params, element_input)

        logger.debug(f"Total calibration sweep process took {time.time() - start_lo_sweep}")
        logger.debug(f"Total calibration process took {time.time() - t_start}")
        return result

    def _set_if_freq(
        self,
        if_freq: FreqType,
        down_mixer_offset: FreqType,
    ) -> None:
        element_name_to_if_freq = {
            self.names.iq_mixer: if_freq,
            self.names.lo_analyzer: -down_mixer_offset,
            self.names.image_analyzer: -if_freq - down_mixer_offset,
        }
        for element_name, freq_to_set in element_name_to_if_freq.items():
            self._set_element_if_freq(element_name, freq_to_set)

    @abc.abstractmethod
    def _set_analyzer_element_outputs_dc_offsets(self, out1: float, out2: float) -> None:
        pass

    @abc.abstractmethod
    def _set_iq_mixer_element_input_dc_offsets(self, i_offset: float, q_offset: float) -> None:
        pass

    @abc.abstractmethod
    def _execute_job(self, compiled: str) -> Union[RunningQmJob, JobApi]:
        pass

    @abc.abstractmethod
    def _set_io_values(self, io1: Optional[Union[int, float]] = None, io2: Optional[Union[int, float]] = None) -> None:
        pass

    @abc.abstractmethod
    def _set_element_if_freq(self, element_name: str, if_freq: FreqType) -> None:
        pass

    @abc.abstractmethod
    def _compile_program(self, _program: Program) -> str:
        pass

    @abc.abstractmethod
    def _get_element_idle_offsets(self, element: str) -> Tuple[float, float, float, float]:
        pass


class OctaveMixerCalibration(OctaveMixerCalibrationBase):
    def __init__(self, client: Octave, quantum_machine: "QuantumMachine"):
        super().__init__(client)
        self._qm = quantum_machine
        if not self._qm._octave_calibration_elements:
            raise NoCalibrationElements(
                "No calibration elements were found, if you want to use the calibration "
                "please open the QM with the flag 'add_calibration_elements_to_config'."
            )
        self._calibration_elements = self._qm._octave_calibration_elements

    def _set_analyzer_element_outputs_dc_offsets(self, out1: float, out2: float) -> None:
        analyzer_elem = self._qm._octave_calibration_elements[self.names.lo_analyzer]
        analyzer_elem.set_input_dc_offset("out1", out1)
        analyzer_elem.set_input_dc_offset("out2", out2)

    def _set_iq_mixer_element_input_dc_offsets(self, i_offset: float, q_offset: float) -> None:
        iq_mixer_elem = self._qm._octave_calibration_elements[self.names.iq_mixer]
        assert isinstance(iq_mixer_elem.input, UpconvertedInput)
        iq_mixer_elem.input.set_output_dc_offset(i_offset=i_offset, q_offset=q_offset)

    def _execute_job(self, compiled: str) -> RunningQmJob:
        pending_job = self._qm.queue.add_compiled(compiled)
        job = pending_job.wait_for_execution()
        return job

    def _set_io_values(self, io1: Optional[Union[int, float]] = None, io2: Optional[Union[int, float]] = None) -> None:
        self._qm.set_io_values(value_1=io1, value_2=io2)

    def _set_element_if_freq(self, element_name: str, if_freq: FreqType) -> None:
        self._qm._octave_calibration_elements[element_name].set_intermediate_frequency(if_freq)

    def _compile_program(self, _program: Program) -> str:
        return self._qm.compile(_program)

    def _get_element_idle_offsets(self, element: str) -> Tuple[float, float, float, float]:
        i_offset = self._qm.get_output_dc_offset_by_element(element, "I")
        q_offset = self._qm.get_output_dc_offset_by_element(element, "Q")
        input1_offset = self._qm.get_input_dc_offset_by_element(self.names.lo_analyzer, "out1")
        input2_offset = self._qm.get_input_dc_offset_by_element(self.names.lo_analyzer, "out2")
        return i_offset, q_offset, input1_offset, input2_offset

    def _restore_previous_state(self, restore_params: _StateRestoreParams, element_input: UpconvertedInput) -> None:
        super()._restore_previous_state(restore_params, element_input)
        element_input.set_output_dc_offset(i_offset=restore_params.i_offset, q_offset=restore_params.q_offset)
        self._set_analyzer_element_outputs_dc_offsets(
            out1=restore_params.input1_offset, out2=restore_params.input2_offset
        )


class NewApiOctaveMixerCalibration(OctaveMixerCalibrationBase):
    def __init__(self, client: Octave, qm_api: "QmApi"):
        super().__init__(client)
        self._qm_api = qm_api
        self._job: Optional[JobApi] = None

    @property
    def job(self) -> JobApi:
        if self._job is None:
            raise RuntimeError("Job not set")
        return self._job

    def _set_analyzer_element_outputs_dc_offsets(self, out1: float, out2: float) -> None:
        # This one is not implemented in the GW, and not so crucial for the calibration, so we remove it for now.
        return
        # analyzer_elem = self.job.elements[self.names.lo_analyzer]
        # analyzer_elem.outputs["out1"].set_dc_offset(out1)
        # analyzer_elem.outputs["out2"].set_dc_offset(out2)

    def _set_iq_mixer_element_input_dc_offsets(self, i_offset: float, q_offset: float) -> None:
        iq_mixer_elem = self.job.elements[self.names.iq_mixer]
        iq_mixer_elem_input = iq_mixer_elem.input
        assert isinstance(iq_mixer_elem_input, MixInputsApi)
        iq_mixer_elem_input.set_dc_offsets(i_offset=i_offset, q_offset=q_offset)

    def _execute_job(self, compiled: str) -> JobApi:
        job = self._qm_api.add_to_queue(compiled)
        job.wait_until("Running", 10.0)
        self._job = job
        return job

    def _set_io_values(self, io1: Optional[Union[int, float]] = None, io2: Optional[Union[int, float]] = None) -> None:
        self.job.set_io_values(io1=io1, io2=io2)

    def _set_element_if_freq(self, element_name: str, if_freq: FreqType) -> None:
        self.job.elements[element_name].set_intermediate_frequency(if_freq)

    def _compile_program(self, _program: Program) -> str:
        return self._qm_api.compile(_program)

    def _get_element_idle_offsets(self, element: str) -> Tuple[float, float, float, float]:
        """This is not necessary as the offsets are set through the job"""
        return 0, 0, 0, 0
